from typing import Any, List

from marshmallow import fields, post_load, utils, missing

import prefect
from prefect.serialization.edge import EdgeSchema
from prefect.serialization.run_config import RunConfigSchema
from prefect.serialization.schedule import ScheduleSchema
from prefect.serialization.storage import StorageSchema
from prefect.serialization.task import ParameterSchema, TaskSchema
from prefect.utilities.serialization import (
    Nested,
    ObjectSchema,
    to_qualified_name,
)


def get_parameters(obj: prefect.Flow, context: dict) -> List:
    if isinstance(obj, prefect.Flow):
        params = obj.parameters()
    else:
        params = utils.get_value(obj, "parameters")

    if params is missing:
        return params

    return sorted(params, key=lambda p: p.slug)


def get_reference_tasks(obj: prefect.Flow, context: dict) -> List:
    if isinstance(obj, prefect.Flow):
        tasks = obj._reference_tasks
    else:
        tasks = utils.get_value(obj, "reference_tasks")

    if tasks is missing:
        return tasks

    return sorted(tasks, key=lambda t: t.slug)


def get_tasks(obj: prefect.Flow, context: dict) -> List:
    if isinstance(obj, prefect.Flow):
        tasks = obj.tasks
    else:
        tasks = utils.get_value(obj, "tasks")

    if tasks is missing:
        return tasks

    return list(sorted(tasks, key=lambda t: t.slug))


def get_edges(obj: prefect.Flow, context: dict) -> List:
    if isinstance(obj, prefect.Flow):
        edges = obj.edges
    else:
        edges = utils.get_value(obj, "edges")

    if edges is missing:
        return edges

    return list(
        sorted(
            edges,
            key=lambda e: (e.upstream_task.slug, e.downstream_task.slug, e.key or ""),
        )
    )


class FlowSchema(ObjectSchema):
    # All nested 'many' types that are stored as a `Set` in the `Flow` must be sorted
    # using a `value_selection_fn` so `Flow.serialized_hash()` is deterministic

    class Meta:
        object_class = lambda: prefect.core.Flow
        exclude_fields = ["type", "parameters"]
        # ordered to make sure Task objects are loaded before Edge objects, due to Task caching
        ordered = True

    project = fields.String(allow_none=True)
    name = fields.String(required=True, allow_none=True)
    version = fields.String(allow_none=True)
    description = fields.String(allow_none=True)
    type = fields.Function(lambda flow: to_qualified_name(type(flow)), lambda x: x)
    schedule = fields.Nested(ScheduleSchema, allow_none=True)
    parameters = Nested(ParameterSchema, value_selection_fn=get_parameters, many=True)
    tasks = Nested(TaskSchema, value_selection_fn=get_tasks, many=True)
    edges = Nested(EdgeSchema, value_selection_fn=get_edges, many=True)
    reference_tasks = Nested(
        TaskSchema, value_selection_fn=get_reference_tasks, many=True, only=["slug"]
    )
    run_config = fields.Nested(RunConfigSchema, allow_none=True)
    storage = fields.Nested(StorageSchema, allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> prefect.core.Flow:
        """
        Flow edges are validated, for example to make sure the keys match Task inputs,
        but because we are deserializing all Tasks as base Tasks, the edge validation will
        fail (base Tasks have no inputs). Therefore we hold back the edges from Flow
        initialization and assign them explicitly.

        Args:
            - data (dict): the deserialized data

        Returns:
            - Flow

        """
        data["validate"] = False
        flow = super().create_object(data)
        return flow



import asyncio
import logging
import uuid
import sys
import weakref
from contextlib import contextmanager
from typing import Any, Callable, Iterator, TYPE_CHECKING, Union, Optional, Dict

from prefect.utilities.compatibility import nullcontext

from prefect import context
from prefect.executors.base import Executor
from prefect.utilities.importtools import import_object

if TYPE_CHECKING:
    from distributed import Future, Event
    from dask.delayed import Delayed
    import concurrent.futures


__any__ = ("DaskExecutor", "LocalDaskExecutor")


def _make_task_key(
    task_name: str = "", task_index: int = None, **kwargs: Any
) -> Optional[str]:
    """A helper for generating a dask task key from fields set in `extra_context`"""
    if task_name:
        suffix = uuid.uuid4().hex
        if task_index is not None:
            return f"{task_name}-{task_index}-{suffix}"
        return f"{task_name}-{suffix}"
    return None


def _maybe_run(event_name: str, fn: Callable, *args: Any, **kwargs: Any) -> Any:
    """Check if the task should run against a `distributed.Event` before
    starting the task. This offers stronger guarantees than distributed's
    current cancellation mechanism, which only cancels pending tasks."""
    import dask
    from distributed import Event, get_client

    try:
        # Explicitly pass in the timeout from dask's config. Some versions of
        # distributed hardcode this rather than using the value from the
        # config.  Can be removed once we bump our min requirements for
        # distributed to >= 2.31.0.
        timeout = dask.config.get("distributed.comm.timeouts.connect")
        event = Event(event_name, client=get_client(timeout=timeout))
        should_run = event.is_set()
    except Exception:
        # Failure to create an event is usually due to connection errors. These
        # are either due to flaky behavior in distributed's comms under high
        # loads, or due to the scheduler shutting down. Either way, the safest
        # course here is to assume we *should* run the task still. If we guess
        # wrong, we're either doing a bit of unnecessary work, or the cluster
        # is shutting down and the task will be cancelled anyway.
        should_run = True

    if should_run:
        return fn(*args, **kwargs)


class DaskExecutor(Executor):
    """
    An executor that runs all functions using the `dask.distributed` scheduler.
    By default a temporary `distributed.LocalCluster` is created (and
    subsequently torn down) within the `start()` contextmanager. To use a
    different cluster class (e.g.
    [`dask_kubernetes.KubeCluster`](https://kubernetes.dask.org/)), you can
    specify `cluster_class`/`cluster_kwargs`.
    Alternatively, if you already have a dask cluster running, you can provide
    the address of the scheduler via the `address` kwarg.
    Note that if you have tasks with tags of the form `"dask-resource:KEY=NUM"`
    they will be parsed and passed as
    [Worker Resources](https://distributed.dask.org/en/latest/resources.html)
    of the form `{"KEY": float(NUM)}` to the Dask Scheduler.
    Args:
        - address (string, optional): address of a currently running dask
            scheduler; if one is not provided, a temporary cluster will be
            created in `executor.start()`.  Defaults to `None`.
        - cluster_class (string or callable, optional): the cluster class to use
            when creating a temporary dask cluster. Can be either the full
            class name (e.g. `"distributed.LocalCluster"`), or the class itself.
        - cluster_kwargs (dict, optional): addtional kwargs to pass to the
           `cluster_class` when creating a temporary dask cluster.
        - adapt_kwargs (dict, optional): additional kwargs to pass to `cluster.adapt`
            when creating a temporary dask cluster. Note that adaptive scaling
            is only enabled if `adapt_kwargs` are provided.
        - client_kwargs (dict, optional): additional kwargs to use when creating a
            [`dask.distributed.Client`](https://distributed.dask.org/en/latest/api.html#client).
        - debug (bool, optional): When running with a local cluster, setting
            `debug=True` will increase dask's logging level, providing
            potentially useful debug info. Defaults to the `debug` value in
            your Prefect configuration.
        - performance_report_path (str, optional): An optional path for the [dask performance
            report](https://distributed.dask.org/en/latest/api.html#distributed.performance_report).
        - disable_cancellation_event (bool, optional): By default, Prefect uses a
            Dask event to allow for better cancellation of task runs. Sometimes this
            can cause strain on the scheduler as each task needs to retrieve a client
            to check the status of the cancellation event. If set to `False`, we will
            skip this check.
        - watch_worker_status (bool, optional): By default, Prefect subscribes to Dask
            worker events and logs when a worker is added or removed. This provides a
            hook for users to extend behavior on worker changes. This setting is `None`
            by default and will be enabled unless `adapt_kwargs` is set, in which case
            it will be disabled. Adaptive clusters often require this feature to be
            disabled as they use the worker status events for scaling and only one
            subscriber is allowed. If you set the value to `True` or `False`, it will be
            respected regardless of the value of `adapt_kwargs`.
    Examples:
    Using a temporary local dask cluster:
    ```python
    executor = DaskExecutor()
    ```
    Using a temporary cluster running elsewhere. Any Dask cluster class should
    work, here we use [dask-cloudprovider](https://cloudprovider.dask.org):
    ```python
    executor = DaskExecutor(
        cluster_class="dask_cloudprovider.FargateCluster",
        cluster_kwargs={
            "image": "prefecthq/prefect:latest",
            "n_workers": 5,
            ...
        },
    )
    ```
    Connecting to an existing dask cluster
    ```python
    executor = DaskExecutor(address="192.0.2.255:8786")
    ```
    """

    def __init__(
        self,
        address: str = None,
        cluster_class: Union[str, Callable] = None,
        cluster_kwargs: dict = None,
        adapt_kwargs: dict = None,
        client_kwargs: dict = None,
        debug: bool = None,
        performance_report_path: str = None,
        disable_cancellation_event: bool = False,
        watch_worker_status: bool = None,
    ):
        if address is None:
            address = context.config.engine.executor.dask.address or None

        if address is not None:
            if cluster_class is not None or cluster_kwargs is not None:
                raise ValueError(
                    "Cannot specify `address` and `cluster_class`/`cluster_kwargs`"
                )
        else:
            if cluster_class is None:
                cluster_class = context.config.engine.executor.dask.cluster_class
            if isinstance(cluster_class, str):
                cluster_class = import_object(cluster_class)
            if cluster_kwargs is None:
                cluster_kwargs = {}
            else:
                cluster_kwargs = cluster_kwargs.copy()

            from distributed.deploy.local import LocalCluster

            if cluster_class == LocalCluster:
                if debug is None:
                    debug = context.config.debug
                cluster_kwargs.setdefault(
                    "silence_logs", logging.CRITICAL if not debug else logging.WARNING
                )

            if adapt_kwargs is None:
                adapt_kwargs = {}

        if client_kwargs is None:
            client_kwargs = {}
        else:
            client_kwargs = client_kwargs.copy()
        client_kwargs.setdefault("set_as_default", False)

        self.address = address
        self.cluster_class = cluster_class
        self.cluster_kwargs = cluster_kwargs
        self.adapt_kwargs = adapt_kwargs
        self.client_kwargs = client_kwargs
        self.disable_cancellation_event = disable_cancellation_event
        self.watch_worker_status = watch_worker_status

        # Runtime attributes
        self.client = None
        # These are coupled - they're either both None, or both non-None.
        # They're used in the case we can't forcibly kill all the dask workers,
        # and need to wait for all the dask tasks to cleanup before exiting.
        self._futures = None  # type: Optional[weakref.WeakSet[Future]]
        self._should_run_event = None  # type: Optional[Event]
        # A ref to a background task subscribing to dask cluster events
        self._watch_dask_events_task = None  # type: Optional[concurrent.futures.Future]

        self.performance_report_path = performance_report_path

        super().__init__()

    @contextmanager
    def start(self) -> Iterator[None]:
        """
        Context manager for initializing execution.
        Creates a `dask.distributed.Client` and yields it.
        """
        if sys.platform != "win32":
            # Fix for https://github.com/dask/distributed/issues/4168
            import multiprocessing.popen_spawn_posix  # noqa
        from distributed import Client, performance_report

        performance_report_context = (
            performance_report(self.performance_report_path)
            if self.performance_report_path
            else nullcontext()
        )

        try:
            if self.address is not None:
                self.logger.info(
                    "Connecting to an existing Dask cluster at %s", self.address
                )
                with Client(self.address, **self.client_kwargs) as client:

                    with performance_report_context:
                        self.client = client
                        try:
                            self._pre_start_yield()
                            yield
                        finally:
                            self._post_start_yield()
            else:
                assert callable(self.cluster_class)  # mypy
                assert isinstance(self.cluster_kwargs, dict)  # mypy
                self.logger.info(
                    "Creating a new Dask cluster with `%s.%s`...",
                    self.cluster_class.__module__,
                    self.cluster_class.__qualname__,
                )
                with self.cluster_class(**self.cluster_kwargs) as cluster:
                    if getattr(cluster, "dashboard_link", None):
                        self.logger.info(
                            "The Dask dashboard is available at %s",
                            cluster.dashboard_link,
                        )
                    if self.adapt_kwargs:
                        cluster.adapt(**self.adapt_kwargs)
                    with Client(cluster, **self.client_kwargs) as client:
                        with performance_report_context:
                            self.client = client
                            try:
                                self._pre_start_yield()
                                yield
                            finally:
                                self._post_start_yield()
        finally:
            self.client = None

    async def on_worker_status_changed(self, op: str, message: dict) -> None:
        """
        This method is triggered when a worker is added or removed from the cluster.
        This method will not be called if `watch_worker_status` is not set.
        We recommend not relying on this method since worker status subscription is used
        by Dask cluster implementations to manage worker state.
        Args:
            - op (str): Either "add" or "remove"
            - message (dict): Information about the event that the scheduler has sent
        """
        if op == "add":
            for worker in message.get("workers", ()):
                self.logger.debug("Worker %s added", worker)
        elif op == "remove":
            self.logger.debug("Worker %s removed", message)

    async def _watch_dask_events(self) -> None:
        scheduler_comm = None
        comm = None
        from distributed.core import rpc

        try:
            scheduler_comm = rpc(
                self.client.scheduler.address,  # type: ignore
                connection_args=self.client.security.get_connection_args("client"),  # type: ignore
            )
            # due to a bug in distributed's inproc comms, letting cancellation
            # bubble up here will kill the listener. wrap with a shield to
            # prevent that.
            comm = await asyncio.shield(scheduler_comm.live_comm())
            await comm.write({"op": "subscribe_worker_status"})
            _ = await comm.read()
            while True:
                try:
                    msgs = await comm.read()
                except OSError:
                    break
                for op, msg in msgs:
                    await self.on_worker_status_changed(op, msg)
        except asyncio.CancelledError:
            pass
        except Exception:
            self.logger.debug(
                "Failure while watching dask worker events", exc_info=True
            )
        finally:
            if comm is not None:
                try:
                    await comm.close()
                except Exception:
                    pass
            if scheduler_comm is not None:
                await scheduler_comm.close_rpc()

    def _pre_start_yield(self) -> None:
        from distributed import Event

        is_inproc = self.client.scheduler.address.startswith("inproc")  # type: ignore
        if (
            self.address is not None or is_inproc
        ) and not self.disable_cancellation_event:
            self._futures = weakref.WeakSet()
            self._should_run_event = Event(
                f"prefect-{uuid.uuid4().hex}", client=self.client
            )
            self._should_run_event.set()

        if self.watch_worker_status is True or (
            self.watch_worker_status is None and not self.adapt_kwargs
        ):
            self._watch_dask_events_task = asyncio.run_coroutine_threadsafe(
                self._watch_dask_events(), self.client.loop.asyncio_loop  # type: ignore
            )

    def _post_start_yield(self) -> None:
        from distributed import wait

        if self._watch_dask_events_task is not None:
            try:
                self._watch_dask_events_task.cancel()
            except Exception:
                pass
            self._watch_dask_events_task = None

        if self._should_run_event is not None:
            # Multipart cleanup, ignoring exceptions in each stage
            # 1.) Stop pending tasks from starting
            try:
                self._should_run_event.clear()
            except Exception:
                pass
            # 2.) Wait for all running tasks to complete
            try:
                futures = [f for f in list(self._futures) if not f.done()]  # type: ignore
                if futures:
                    self.logger.info(
                        "Stopping executor, waiting for %d active tasks to complete",
                        len(futures),
                    )
                    wait(futures)
            except Exception:
                pass
        self._should_run_event = None
        self._futures = None

    def _prep_dask_kwargs(self, extra_context: dict = None) -> dict:
        if extra_context is None:
            extra_context = {}

        dask_kwargs = {"pure": False}  # type: dict

        # set a key for the dask scheduler UI
        key = _make_task_key(**extra_context)
        if key is not None:
            dask_kwargs["key"] = key

        # infer from context if dask resources are being utilized
        task_tags = extra_context.get("task_tags", [])
        dask_resource_tags = [
            tag for tag in task_tags if tag.lower().startswith("dask-resource")
        ]
        if dask_resource_tags:
            resources = {}
            for tag in dask_resource_tags:
                prefix, val = tag.split("=")
                resources.update({prefix.split(":")[1]: float(val)})
            dask_kwargs.update(resources=resources)

        return dask_kwargs

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        state.update(
            {
                k: None
                for k in [
                    "client",
                    "_futures",
                    "_should_run_event",
                    "_watch_dask_events_task",
                ]
            }
        )
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)

    def submit(
        self, fn: Callable, *args: Any, extra_context: dict = None, **kwargs: Any
    ) -> "Future":
        """
        Submit a function to the executor for execution. Returns a Future object.
        Args:
            - fn (Callable): function that is being submitted for execution
            - *args (Any): arguments to be passed to `fn`
            - extra_context (dict, optional): an optional dictionary with extra information
                about the submitted task
            - **kwargs (Any): keyword arguments to be passed to `fn`
        Returns:
            - Future: a Future-like object that represents the computation of `fn(*args, **kwargs)`
        """
        if self.client is None:
            raise ValueError("This executor has not been started.")

        kwargs.update(self._prep_dask_kwargs(extra_context))
        if self._should_run_event is None:
            fut = self.client.submit(fn, *args, **kwargs)
        else:
            fut = self.client.submit(
                _maybe_run, self._should_run_event.name, fn, *args, **kwargs
            )
            self._futures.add(fut)
        return fut

    def wait(self, futures: Any) -> Any:
        """
        Resolves the Future objects to their values. Blocks until the computation is complete.
        Args:
            - futures (Any): single or iterable of future-like objects to compute
        Returns:
            - Any: an iterable of resolved futures with similar shape to the input
        """
        if self.client is None:
            raise ValueError("This executor has not been started.")

        return self.client.gather(futures)

    @property
    def performance_report(self) -> str:
        """The performance report html string."""
        if self.performance_report_path is None:
            self.logger.warning(
                "Executor was not configured to generate performance report"
            )
            return ""
        self.logger.debug(
            f"Retreiving dask performance report from {self.performance_report_path!r}"
        )
        try:
            with open(self.performance_report_path, "r", encoding="utf-8") as f:
                report = f.read()
                return report
        except Exception as exc:
            self.logger.error(
                f"Failed to get dask performance report with exception {exc}"
            )
            return ""


def _multiprocessing_pool_initializer() -> None:
    """Initialize a process used in a `concurrent.futures.ProcessPoolExecutor`.
    Ensures the standard atexit handlers are run."""
    import signal

    signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())


class LocalDaskExecutor(Executor):
    """
    An executor that runs all functions locally using `dask` and a configurable
    dask scheduler.
    Args:
        - scheduler (str): The local dask scheduler to use; common options are
            "threads", "processes", and "synchronous".  Defaults to "threads".
        - **kwargs (Any): Additional keyword arguments to pass to dask config
    """

    def __init__(self, scheduler: str = "threads", **kwargs: Any):
        self.scheduler = self._normalize_scheduler(scheduler)
        self.dask_config = kwargs
        self._pool = None  # type: Optional[concurrent.futures.Executor]
        super().__init__()

    @staticmethod
    def _normalize_scheduler(scheduler: str) -> str:
        scheduler = scheduler.lower()
        if scheduler in ("threads", "threading"):
            return "threads"
        elif scheduler in ("processes", "multiprocessing"):
            return "processes"
        elif scheduler in ("sync", "synchronous", "single-threaded"):
            return "synchronous"
        else:
            raise ValueError(f"Unknown scheduler {scheduler!r}")

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        state["_pool"] = None
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)

    def _interrupt_pool(self) -> None:
        """Interrupt all tasks in the backing `pool`, if any."""
        if self._pool is None:
            return

        # Shutdown the pool
        self._pool.shutdown(wait=False)

        if self.scheduler == "threads":
            # `ThreadPoolExecutor.shutdown()` doesn't stop running tasks, only
            # prevents new tasks from running. In CPython we can attempt to
            # raise an exception in all threads. This exception will be raised
            # the next time the task does something with the Python api.
            # However, if the task is currently blocked in a c extension, it
            # will not immediately be interrupted. There isn't a good way
            # around this unfortunately.
            import platform

            if platform.python_implementation() != "CPython":
                self.logger.warning(
                    "Interrupting a running threadpool is only supported in CPython, "
                    "all currently running tasks will continue to completion"
                )
                return

            self.logger.info("Attempting to interrupt and cancel all running tasks...")

            import sys
            import ctypes

            # signature of this method changed in python 3.7
            if sys.version_info >= (3, 7):
                id_type = ctypes.c_ulong
            else:
                id_type = ctypes.c_long

            for t in self._pool._threads:  # type: ignore
                ctypes.pythonapi.PyThreadState_SetAsyncExc(
                    id_type(t.ident), ctypes.py_object(KeyboardInterrupt)
                )

    @contextmanager
    def start(self) -> Iterator:
        """Context manager for initializing execution."""
        # import dask here to reduce prefect import times
        import dask.config
        from dask.callbacks import Callback
        from dask.system import CPU_COUNT

        class PrefectCallback(Callback):
            def __init__(self):  # type: ignore
                self.cache = {}

            def _start(self, dsk):  # type: ignore
                overlap = set(dsk) & set(self.cache)
                for key in overlap:
                    dsk[key] = self.cache[key]

            def _posttask(self, key, value, dsk, state, id):  # type: ignore
                self.cache[key] = value

        with PrefectCallback(), dask.config.set(**self.dask_config):
            if self.scheduler == "synchronous":
                self._pool = None
            else:
                num_workers = dask.config.get("num_workers", None) or CPU_COUNT
                if self.scheduler == "threads":
                    from concurrent.futures import ThreadPoolExecutor

                    self._pool = ThreadPoolExecutor(num_workers)
                else:
                    from concurrent.futures import ProcessPoolExecutor

                    self._pool = ProcessPoolExecutor(
                        num_workers, initializer=_multiprocessing_pool_initializer
                    )
            try:
                exiting_early = False
                yield
            except BaseException:
                exiting_early = True
                raise
            finally:
                if self._pool is not None:
                    if exiting_early:
                        self._interrupt_pool()
                    else:
                        self._pool.shutdown(wait=True)
                    self._pool = None

    def submit(
        self, fn: Callable, *args: Any, extra_context: dict = None, **kwargs: Any
    ) -> "Delayed":
        """
        Submit a function to the executor for execution. Returns a `dask.delayed` object.
        Args:
            - fn (Callable): function that is being submitted for execution
            - *args (Any): arguments to be passed to `fn`
            - extra_context (dict, optional): an optional dictionary with extra
                information about the submitted task
            - **kwargs (Any): keyword arguments to be passed to `fn`
        Returns:
            - dask.delayed: a `dask.delayed` object that represents the
                computation of `fn(*args, **kwargs)`
        """
        # import dask here to reduce prefect import times
        import dask

        extra_kwargs = {}
        key = _make_task_key(**(extra_context or {}))
        if key is not None:
            extra_kwargs["dask_key_name"] = key
        return dask.delayed(fn, pure=False)(*args, **kwargs, **extra_kwargs)  # type: ignore

    def wait(self, futures: Any) -> Any:
        """
        Resolves a (potentially nested) collection of `dask.delayed` object to
        its values. Blocks until the computation is complete.
        Args:
            - futures (Any): iterable of `dask.delayed` objects to compute
        Returns:
            - Any: an iterable of resolved futures
        """
        # import dask here to reduce prefect import times
        import dask

        config: Dict[str, Any] = {}  # Extra config options for dask

        # dask's multiprocessing scheduler hardcodes task fusion in a way
        # that's not exposed via a `compute` kwarg. Until that's fixed, we
        # disable fusion globally for the multiprocessing scheduler only.
        # Since multiprocessing tasks execute in a remote process, this
        # shouldn't affect user code.
        if self.scheduler == "processes":

            @contextmanager
            def patch() -> Iterator[None]:
                # Patch around https://github.com/PrefectHQ/prefect/issues/4086
                # We can remove this after we drop support for dask 2021.02.0
                from dask.optimization import cull

                def cull2(dsk, keys):  # type: ignore
                    return cull(dsk if type(dsk) is dict else dict(dsk), keys)

                dask.multiprocessing.cull = cull2
                try:
                    yield
                finally:
                    dask.multiprocessing.cull = cull

            config["optimization.fuse.active"] = False
        else:

            @contextmanager
            def patch() -> Iterator[None]:
                yield

        # Patch around https://github.com/PrefectHQ/prefect/issues/4537
        # dask >= 2021.04.0 adds task submission batching in the multiprocessing case
        # to offset performance concerns from switching to a concurrent.Futures based
        # process pool. Since we are using our own pool to handle cancellation robustly,
        # we do not gain anything from the batched submission and the default task
        # submission batch size of 6 makes users think their flows will not run in
        # parallel. When/if we switch to using the new futures based multiprocessing in
        # dask, we should remove this to retain performance
        if self.scheduler == "processes":
            config["chunksize"] = 1

        with patch(), dask.config.set(config):
            return dask.compute(
                futures, scheduler=self.scheduler, pool=self._pool, optimize_graph=False
            )[0]



import cloudpickle
import contextlib
import copy
import itertools
import logging
import multiprocessing
import os
import signal
import subprocess
import sys
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as FutureTimeout
from contextlib import contextmanager
from functools import wraps
from logging import Logger
from queue import Empty

import prefect
from prefect import config
from prefect.client import Client
from prefect.configuration import to_environment_variables
from prefect.exceptions import TaskTimeoutSignal, PrefectSignal
from prefect.utilities.logging import get_logger
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Union,
    Sequence,
    Mapping,
    Iterator,
)

if TYPE_CHECKING:
    import prefect.engine.runner
    import prefect.engine.state
    from prefect.core.edge import Edge
    from prefect.core.task import Task
    from prefect.engine.state import State


StateList = Union["State", List["State"]]

FLOWRUNNER_LOGGER = get_logger("FlowRunner")


def run_with_heartbeat(
    runner_method: Callable[..., "prefect.engine.state.State"]
) -> Callable[..., "prefect.engine.state.State"]:
    """
    Utility decorator for running class methods with a heartbeat.  The class should implement
    `self._heartbeat` with no arguments.
    """

    @wraps(runner_method)
    def inner(
        self: "prefect.engine.cloud.CloudFlowRunner", *args: Any, **kwargs: Any
    ) -> "prefect.engine.state.State":

        try:
            use_heartbeat = self._heartbeat()
        except Exception:
            use_heartbeat = False
            logger = self.logger
            logger.exception(
                "Heartbeat process is misconfigured.  This could result in a zombie run.",
                exc_info=True,
            )

        if not use_heartbeat or (prefect.context.config.cloud.heartbeat_mode == "off"):
            configured_heartbeat = no_heartbeat()
        elif prefect.context.config.cloud.heartbeat_mode == "thread":
            configured_heartbeat = threaded_heartbeat(self.flow_run_id)
        elif prefect.context.config.cloud.heartbeat_mode == "process":
            configured_heartbeat = subprocess_heartbeat(self.heartbeat_cmd, self.logger)
        # because the threaded heartbeat mode is experimental and may change in the future,
        # let's not catch configuration error with `else` -- stale configuration should break tests

        with configured_heartbeat:
            return runner_method(self, *args, **kwargs)

    return inner


@contextlib.contextmanager
def no_heartbeat() -> Iterator[None]:
    # contextlib.nullcontext was introduced in 3.7
    yield


@contextlib.contextmanager
def threaded_heartbeat(flow_run_id: str, num: int = None) -> Iterator[None]:
    try:
        HEARTBEAT_STOP_EVENT = threading.Event()
        heartbeat = HeartbeatThread(HEARTBEAT_STOP_EVENT, flow_run_id, num=None)
        heartbeat.start()
        yield
    finally:
        HEARTBEAT_STOP_EVENT.set()


@contextlib.contextmanager
def subprocess_heartbeat(heartbeat_cmd: List[str], logger: Logger) -> Iterator[None]:
    p = None
    try:
        # we use Popen + a prefect CLI for a few reasons:
        # - using threads would interfere with the task; for example, a task
        #   which does not release the GIL would prevent the heartbeat thread from
        #   firing
        # - using multiprocessing.Process would release the GIL but a subprocess
        #   cannot be spawned from a daemonic subprocess, and Dask sometimes will
        #   submit tasks to run within daemonic subprocesses
        current_env = dict(os.environ).copy()
        current_env.update(
            to_environment_variables(
                prefect.context.config,
                include={
                    "cloud.api_key",
                    "cloud.tenant_id",
                    "cloud.api",
                },
            )
        )
        clean_env = {k: v for k, v in current_env.items() if v is not None}
        p = subprocess.Popen(
            heartbeat_cmd,
            env=clean_env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        yield

    except Exception:
        logger.exception(
            "Heartbeat process failed to start.  This could result in a zombie run."
        )

    finally:
        if p is not None:
            exit_code = p.poll()
            if exit_code is not None:
                logger.error(f"Heartbeat process died with exit code {exit_code}")
            p.kill()
            p.wait()


class HeartbeatThread(threading.Thread):
    def __init__(
        self: "HeartbeatThread",
        stop_event: "threading.Event",
        flow_run_id: str,
        num: int = None,
    ) -> None:
        threading.Thread.__init__(self)
        # 'daemonizes' the thread, so it will terminate when all non-daemonized threads have finished
        self.daemon = True
        self.flow_run_id = flow_run_id
        self.num = num
        self.stop_event = stop_event

    def run(self) -> None:
        logger = get_logger("threaded_heartbeat")
        client = Client()
        iter_count = 0
        with prefect.context(
            {"flow_run_id": self.flow_run_id, "running_with_backend": True}
        ):
            with log_heartbeat_failure(logger):
                while iter_count < (self.num or 1) and (
                    self.stop_event.is_set() is False
                ):
                    send_heartbeat(self.flow_run_id, client, logger)
                    iter_count += 1 if self.num else 0
                    self.stop_event.wait(timeout=config.cloud.heartbeat_interval)


def send_heartbeat(
    flow_run_id: str, client: "prefect.client.Client", logger: "logging.Logger"
) -> None:
    try:  # Ignore (but log) client exceptions
        client.update_flow_run_heartbeat(flow_run_id)
    except Exception as exc:
        logger.error(
            f"Failed to send heartbeat with exception: {exc!r}",
            exc_info=True,
        )


@contextmanager
def log_heartbeat_failure(
    logger: "logging.Logger",
) -> Iterator[None]:
    try:
        yield
    except BaseException as exc:
        logger.error(
            f"Heartbeat process encountered terminal exception: {exc!r}",
            exc_info=True,
        )
        raise


def run_with_thread_timeout(
    fn: Callable,
    args: Sequence = (),
    kwargs: Mapping = None,
    timeout: int = None,
    logger: Logger = None,
    name: str = None,
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by setting a `signal` alarm on a timer. Must be run in the main thread.

    Args:
        - fn (callable): the function to execute
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - timeout (int): the length of time to allow for execution before raising a
            `TaskTimeoutSignal`, represented as an integer in seconds
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.executors.run_with_thread_timeout` namespace will be created.
        - name (str): an optional name to attach to logs for this function run, defaults
            to the name of the given function. Provides an interface for passing task
            names for logs.

    Returns:
        - the result of `fn(*args, **kwargs)`

    Raises:
        - TaskTimeoutSignal: if function execution exceeds the allowed timeout
        - ValueError: if run from outside the main thread
    """
    logger = logger or get_logger()
    name = name or f"Function '{fn.__name__}'"
    kwargs = kwargs or {}

    if timeout is None:
        return fn(*args, **kwargs)

    def error_handler(signum, frame):  # type: ignore
        raise TaskTimeoutSignal("Execution timed out.")

    try:
        # Set the signal handler for alarms
        signal.signal(signal.SIGALRM, error_handler)
        # Raise the alarm if `timeout` seconds pass
        logger.debug(f"{name}: Sending alarm with {timeout}s timeout...")
        signal.alarm(timeout)
        logger.debug(f"{name}: Executing function in main thread...")
        return fn(*args, **kwargs)
    finally:
        signal.alarm(0)


def multiprocessing_safe_run_and_retrieve(
    queue: multiprocessing.Queue,
    payload: bytes,
) -> None:
    """
    Gets the return value from a function and puts it in a multiprocessing-safe
    container. Helper function for `run_with_multiprocess_timeout`, must be defined
    top-level so it can be pickled and sent to `multiprocessing.Process`

    Passing the payload serialized allows us to escape the limitations of the python
    native pickler which will fail on tasks defined in scripts because of name
    mismatches. Whilst this particular example only affects the `func` arg, any of the
    others could be affected by other pickle limitations as well.

    Args:
        - queue (multiprocessing.Queue): The queue to pass the resulting payload to
        - payload (bytes): A serialized dictionary containing the data required to run
            the function. Should be serialized with `cloudpickle.dumps`
            Expects the following keys:
            - fn (Callable): The function to call
            - args (list): Positional argument values to call the function with
            - kwargs (Mapping): Keyword arguments to call the function with
            - context (dict): The prefect context dictionary to use during execution
            - name (str): an optional name to attach to logs for this function run,
                defaults to the name of the given function. Provides an interface for
                passing task names for logs.
            - logger (Logger): the logger to use
    """
    request = cloudpickle.loads(payload)

    fn: Callable = request["fn"]
    context: dict = request.get("context", {})
    args: Sequence = request.get("args", [])
    kwargs: Mapping = request.get("kwargs", {})
    name: str = request.get("name", f"Function '{fn.__name__}'")
    logger: Logger = request.get("logger") or get_logger()

    try:
        with prefect.context(context):
            logger.debug(f"{name}: Executing...")
            return_val = fn(*args, **kwargs)
            logger.debug(f"{name}: Execution successful.")
    except PrefectSignal as exc:
        # We need to ensure PrefectSignals are captured and put in the
        # queue correctly, otherwise no value is returned and
        # the task runner assumes execution has timed out
        return_val = exc
        logger.error(
            f"{name}: Encountered a PrefectSignal {type(exc).__name__}, "
            f"returning details as a result..."
        )
    except Exception as exc:
        return_val = exc
        logger.error(
            f"{name}: Encountered a {type(exc).__name__}, "
            f"returning details as a result..."
        )

    try:
        logger.debug(
            "%s: Pickling value of size %s...", name, sys.getsizeof(return_val, -1)
        )  # Only calculate the size if debug logs are enabled
        pickled_val = cloudpickle.dumps(return_val)
        logger.debug(f"{name}: Pickling successful!")
    except Exception as exc:
        err_msg = (
            f"Failed to pickle result of type {type(return_val).__name__!r} with "
            f'exception: "{type(exc).__name__}: {str(exc)}". This timeout handler "'
            "requires your function return value to be serializable with `cloudpickle`."
        )
        logger.error(f"{name}: {err_msg}")
        pickled_val = cloudpickle.dumps(RuntimeError(err_msg))

    logger.debug(f"{name}: Passing result back to main process...")

    try:
        queue.put(pickled_val)
    except Exception:
        logger.error(
            f"{name}: Failed to put result in queue to main process!",
            exc_info=True,
        )
        raise


def run_with_multiprocess_timeout(
    fn: Callable,
    args: Sequence = (),
    kwargs: Mapping = None,
    timeout: int = None,
    logger: Logger = None,
    name: str = None,
) -> Any:
    """
    Helper function for implementing timeouts on function executions.

    Implemented by spawning a new multiprocess.Process() and using a queue to pass
    the result back. The result is retrieved from the queue with a timeout.

    Args:
        - fn (callable): the function to execute
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - timeout (int): the length of time to allow for execution before raising a
            `TaskTimeoutSignal`, represented as an integer in seconds
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.` namespace will be created.
        - name (str): an optional name to attach to logs for this function run, defaults
            to the name of the given function. Provides an interface for passing task
            names for logs.

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - Exception: Any user errors within the subprocess will be pickled and reraised
        - AssertionError: if run from a daemonic process
        - TaskTimeoutSignal: if function execution exceeds the allowed timeout
    """
    logger = logger or get_logger()
    name = name or f"Function '{fn.__name__}'"
    kwargs = kwargs or {}

    if timeout is None:
        return fn(*args, **kwargs)

    spawn_mp = multiprocessing.get_context("spawn")

    # Create a queue to pass the function return value back
    queue = spawn_mp.Queue()  # type: multiprocessing.Queue

    # Set internal kwargs for the helper function
    request = {
        "fn": fn,
        "args": args,
        "kwargs": kwargs,
        "context": prefect.context.to_dict(),
        "name": name,
        "logger": logger,
    }
    payload = cloudpickle.dumps(request)

    run_process = spawn_mp.Process(
        target=multiprocessing_safe_run_and_retrieve,
        args=(queue, payload),
    )
    logger.debug(f"{name}: Sending execution to a new process...")
    run_process.start()
    logger.debug(f"{name}: Waiting for process to return with {timeout}s timeout...")

    # Pull the data from the queue. If empty, the function did not finish before
    # the timeout
    try:
        pickled_result = queue.get(block=True, timeout=timeout)
        logger.debug(f"{name}: Result received from subprocess, unpickling...")
        result = cloudpickle.loads(pickled_result)
        if isinstance(result, (Exception, PrefectSignal)):
            raise result
        return result
    except Empty:
        logger.debug(f"{name}: No result returned within the timeout period!")
        raise TaskTimeoutSignal(f"Execution timed out for {name}.")
    finally:
        # Do not let the process dangle
        run_process.join(0.1)
        run_process.terminate()


def run_task_with_timeout(
    task: "Task",
    args: Sequence = (),
    kwargs: Mapping = None,
    logger: Logger = None,
) -> Any:
    """
    Helper function for implementing timeouts on task executions.

    The exact implementation varies depending on whether this function is being
    run in the main thread or a non-daemonic subprocess.  If this is run from a
    daemonic subprocess or on Windows, the task is run in a `ThreadPoolExecutor`
    and only a soft timeout is enforced, meaning a `TaskTimeoutSignal` is raised at the
    appropriate time but the task continues running in the background.

    The task is passed instead of a function so we can give better logs and messages.
    If you need to run generic functions with timeout handlers,
    `run_with_thread_timeout` or `run_with_multiprocess_timeout` can be called directly

    Args:
        - task (Task): the task to execute
            `task.timeout` specifies the number of seconds to allow `task.run` to run
            for before terminating
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.run_task_with_timeout_handler` namespace will be created.

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - TaskTimeoutSignal: if function execution exceeds the allowed timeout
    """
    logger = logger or get_logger()
    name = prefect.context.get("task_full_name", task.name)
    kwargs = kwargs or {}

    # if no timeout, just run the function
    if task.timeout is None:
        return task.run(*args, **kwargs)  # type: ignore

    # if we are running the main thread, use a signal to stop execution at the
    # appropriate time; else if we are running in a non-daemonic process, spawn
    # a subprocess to kill at the appropriate time
    if not sys.platform.startswith("win"):

        if threading.current_thread() is threading.main_thread():
            # This case is typically encountered when using a non-parallel or local
            # multiprocess scheduler because then each worker is in the main
            # thread
            logger.debug(f"Task '{name}': Attaching thread based timeout handler...")
            return run_with_thread_timeout(
                task.run,
                args,
                kwargs,
                timeout=task.timeout,
                logger=logger,
                name=f"Task '{name}'",
            )

        elif multiprocessing.current_process().daemon is False:
            # This case is typically encountered when using a multithread distributed
            # executor
            logger.debug(f"Task '{name}': Attaching process based timeout handler...")
            return run_with_multiprocess_timeout(
                task.run,
                args,
                kwargs,
                timeout=task.timeout,
                logger=logger,
                name=f"Task '{name}'",
            )

        # We are in a daemonic process and cannot enforce a timeout
        # This case is typically encountered when using a multiprocess distributed
        # executor
        soft_timeout_reason = "in a daemonic subprocess"
    else:
        # We are in windows and cannot enforce a timeout
        soft_timeout_reason = "on Windows"

    msg = (
        f"This task is running {soft_timeout_reason}; "
        "consequently Prefect can only enforce a soft timeout limit, i.e., "
        "if your Task reaches its timeout limit it will enter a TimedOut state "
        "but continue running in the background."
    )

    logger.debug(
        f"Task '{name}': Falling back to daemonic soft limit timeout handler because "
        f"we are running {soft_timeout_reason}."
    )
    warnings.warn(msg, stacklevel=2)
    executor = ThreadPoolExecutor()

    def run_with_ctx(context: dict) -> Any:
        with prefect.context(context):
            return task.run(*args, **kwargs)  # type: ignore

    # Run the function in the background and then retrieve its result with a timeout
    fut = executor.submit(run_with_ctx, prefect.context.to_dict())

    try:
        return fut.result(timeout=task.timeout)
    except FutureTimeout as exc:
        raise TaskTimeoutSignal(
            f"Execution timed out but was executed {soft_timeout_reason} and will "
            "continue to run in the background."
        ) from exc


class RecursiveCall(Exception):
    def __init__(self, func: Callable, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs


def tail_recursive(func: Callable) -> Callable:
    """
    Helper function to facilitate tail recursion of the wrapped function.

    This allows for recursion with unlimited depth since a stack is not allocated for
    each "nested" call. Note: instead of calling the target function in question, a
    `RecursiveCall` exception must be raised instead.

    Args:
        - fn (callable): the function to execute

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - RecursionError: if a recursive "call" (raised exception) is made with a function that is
            not decorated with `tail_recursive` decorator.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        while True:
            try:
                return func(*args, **kwargs)
            except RecursiveCall as exc:
                try:
                    call_func = getattr(exc.func, "__wrapped_func__")
                except AttributeError as attr_error:
                    raise RecursionError(
                        "function has not been wrapped to provide tail recursion (func={})".format(
                            exc.func
                        )
                    ) from attr_error

                # there may be multiple nested recursive calls, we should only
                # respond to calls for the wrapped function explicitly,
                # otherwise allow the call to continue to propagate
                if call_func != func:
                    raise exc
                args = exc.args
                kwargs = exc.kwargs
                continue

    setattr(wrapper, "__wrapped_func__", func)
    return wrapper


def prepare_upstream_states_for_mapping(
    state: "State",
    upstream_states: "Dict[Edge, State]",
    mapped_children: "Dict[Task, list]",
    executor: "prefect.executors.Executor",
) -> list:
    """
    If the task is being mapped, submits children tasks for execution. Returns a `Mapped` state.

    Args:
        - state (State): the parent task's current state
        - upstream_states (Dict[Edge, State]): the upstream states to this task
        - mapped_children (Dict[Task, List[State]]): any mapped children upstream of this task

    Returns:
        - List: a restructured list of upstream states correponding to each new mapped child task
    """

    # if the current state is failed / skipped or otherwise
    # in a state that signifies we should not continue with mapping,
    # we return an empty list
    if state.is_pending() or state.is_failed() or state.is_skipped():
        return []

    map_upstream_states = []

    # copy the mapped children dict to avoid mutating the global one
    # when we flatten mapped children
    mapped_children = mapped_children.copy()

    # we don't know how long the iterables are, but we want to iterate until we reach
    # the end of the shortest one
    counter = itertools.count()

    # preprocessing
    for edge, upstream_state in upstream_states.items():

        # ensure we are working with populated result objects
        if edge.key in state.cached_inputs:
            upstream_state._result = state.cached_inputs[edge.key]

        # if the upstream was mapped and the edge is flattened, we need
        # to process the mapped children (which could be futures) into a
        # flat structure
        if upstream_state.is_mapped() and edge.flattened:
            mapped_children[edge.upstream_task] = flatten_mapped_children(
                mapped_children=mapped_children[edge.upstream_task], executor=executor
            )

    # infinite loop, if upstream_states has any entries
    while True and upstream_states:
        i = next(counter)
        states = {}
        try:

            for edge, upstream_state in upstream_states.items():

                # if the edge is not mapped over, then we take its state
                if not edge.mapped:
                    states[edge] = upstream_state

                # if the edge is mapped and the upstream state is Mapped, then we are mapping
                # over a mapped task. In this case, we take the appropriately-indexed upstream
                # state from the upstream tasks's `Mapped.map_states` array.
                # Note that these "states" might actually be futures at this time; we aren't
                # blocking until they finish.
                elif edge.mapped and upstream_state.is_mapped():
                    states[edge] = mapped_children[edge.upstream_task][i]  # type: ignore

                # Otherwise, we are mapping over the result of a "vanilla" task. In this
                # case, we create a copy of the upstream state but set the result to the
                # appropriately-indexed item from the upstream task's `State.result`
                # array.
                else:
                    states[edge] = copy.copy(upstream_state)

                    # if the current state is already Mapped, then we might be executing
                    # a re-run of the mapping pipeline. In that case, the upstream states
                    # might not have `result` attributes.
                    # Therefore, we only try to get a result if EITHER this task's
                    # state is not already mapped OR the upstream result is not None.
                    if (
                        not state.is_mapped()
                        or upstream_state._result != prefect.engine.result.NoResult
                    ):
                        if not hasattr(upstream_state.result, "__getitem__"):
                            value = None
                        else:
                            value = upstream_state.result[i]
                        upstream_result = upstream_state._result.from_value(value)  # type: ignore
                        states[edge].result = upstream_result
                        if state.map_states and i >= len(state.map_states):  # type: ignore
                            raise IndexError()
                    elif state.is_mapped():
                        if i >= len(state.map_states):  # type: ignore
                            raise IndexError()

            # only add this iteration if we made it through all iterables
            map_upstream_states.append(states)

        # index error means we reached the end of the shortest iterable
        except IndexError:
            break

    return map_upstream_states


def _should_flatten(state: "State") -> bool:
    return state.is_successful() and not state.is_skipped()


def _can_flatten(state: "State") -> bool:
    return hasattr(state.result, "__len__") and hasattr(state.result, "__getitem__")


def _build_flattened_state(state: "State", index: int) -> "State":
    """Helper function for `flatten_upstream_state`"""
    if not _should_flatten(state):
        return state

    new_state = copy.copy(state)
    if not _can_flatten(state):
        message = (
            "`flatten` was used on upstream task that did not return an iterable. "
            "The value will be passed downstream unmodified."
        )
        FLOWRUNNER_LOGGER.warning(message)
        new_state.result = state._result.from_value(state.result)  # type: ignore
    else:
        new_state.result = state._result.from_value(state.result[index])  # type: ignore

    return new_state


def flatten_upstream_state(upstream_state: "State") -> "State":
    """
    Given an upstream state, returns its result as a flattened list. If
    flattening fails, the object is returned unmodified.
    """
    try:
        # attempt to unnest
        flattened_result = [y for x in upstream_state.result for y in x]
    except TypeError:
        return upstream_state

    new_state = copy.copy(upstream_state)
    new_state.result = new_state._result.from_value(flattened_result)  # type: ignore
    return new_state


def flatten_mapped_children(
    mapped_children: List["State"],
    executor: "prefect.executors.Executor",
) -> List["State"]:
    counts = executor.wait(
        [
            executor.submit(lambda c: len(c._result.value) if _can_flatten(c) else 1, c)
            for c in mapped_children
        ]
    )
    new_states = []

    for child, count in zip(mapped_children, counts):
        new_states.append(
            [executor.submit(_build_flattened_state, child, i) for i in range(count)]
        )

    flattened_states = [i for s in new_states for i in s]
    return flattened_states



import time

import re
import requests

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


MIN_WAIT_TIME, DEFAULT_WAIT_TIME = 5, 60


class CensusSyncTask(Task):
    """
    Task for running Census connector sync jobs.

    This task assumes the user has a Census sync already configured and is attempting to orchestrate the
    sync using Prefect task to send a post to the API within a prefect flow. Copy and paste from the api
    trigger section on the configuration page in the `api_trigger` param to set a default sync.

    Args:
        - api_trigger (str, optional): Default sync to trigger, if none is specified in `run`. The API
            trigger URL for a sync can be found on sync's configuration page
            (https://app.getcensus.com/syncs/{sync_id}/configuration) under Sync Triggers > API.
        - **kwargs (dict, optional): Additional kwargs to pass to the base Task constructor.

    Example:
        Trigger a Census sync with an `api_trigger` stored in a Prefect secret:
        ```python
        from prefect import Flow
        from prefect.tasks.census import CensusSyncTask
        from prefect.tasks.secrets import PrefectSecret

        sync_census = CensusSyncTask()

        with Flow("Run Census Sync") as flow:
            api_trigger = PrefectSecret('CENSUS_API_TRIGGER')
            results = sync_census(api_trigger=api_trigger)
        ```
    """

    def __init__(self, api_trigger=None, **kwargs):
        if api_trigger:
            self.check_invalid_api(api_trigger)
        self.api_trigger = api_trigger
        super().__init__(**kwargs)

    @defaults_from_attrs("api_trigger")
    def run(
        self, api_trigger: str, poll_status_every_n_seconds: int = DEFAULT_WAIT_TIME
    ) -> dict:
        """
        Task run method for Census syncs.

        An invocation of `run` will attempt to start a sync job for the specified `api_trigger`. `run`
        will poll Census for the sync status, and will only complete when the sync has completed or
        when it receives an error status code from the trigger API call.

        Args:
            - api_trigger (str): Ff not specified in run, it will pull from the default for the
                CensusSyncTask constructor. The API trigger URL for a sync can be found on sync's
                configuration page (https://app.getcensus.com/syncs/{sync_id}/configuration) under
                Sync Triggers > API.
            - poll_status_every_n_seconds (int, optional): This task polls the Census API for the sync's
                status. If provided, this value will override the default polling time of
                60 seconds and it has a minimum wait time of 5 seconds. Keyword argument.

        Returns:
            - dict: Dictionary of statistics returned by Census on the specified sync in
                following structure:
                ```python
                {
                    'error_message': None / str,
                    'records_failed': int / None,
                    'records_invalid': int / None,
                    'records_processed': int / None,
                    'records_updated': int / None,
                    'status': 'completed'/'working'/'failed'
                }
                ```
        """

        if not api_trigger:
            raise ValueError(
                """Value for parameter `api_trigger` must be provided. See Census sync
                                configuration page."""
            )

        confirmed_pattern = self.check_invalid_api(api_trigger)

        secret, sync_id = confirmed_pattern.groups()
        response = requests.post(api_trigger)
        response.raise_for_status()

        sleep_time = max(MIN_WAIT_TIME, poll_status_every_n_seconds)

        self.logger.debug(
            f"Started Census sync {sync_id}, sleep time set to {sleep_time} seconds."
        )

        sync_run_id = response.json()["data"]["sync_run_id"]
        sr_url = f"https://bearer:secret-token:{secret}@app.getcensus.com/api/v1/sync_runs/{sync_run_id}"
        log_url = f"https://app.getcensus.com/syncs/{sync_id}/sync-history"

        result = {}

        start_time = time.time()
        while True:
            time.sleep(sleep_time)
            status_response = requests.get(sr_url)
            response_dict = status_response.json()
            if status_response.status_code != 200 or "data" not in response_dict.keys():
                raise ValueError(
                    f"Getting status of sync failed, please visit Census Logs at {log_url} to see more."
                )
            result = response_dict["data"]
            status = result["status"]
            if status == "working":
                self.logger.debug(
                    f"Sync {sync_id} still running after {round(time.time()-start_time, 2)} seconds."
                )
                continue
            break

        self.logger.debug(
            f"Sync {sync_id} has finished running after {round(time.time()-start_time, 2)} seconds."
        )

        self.logger.info(f"View sync details here: {log_url}.")

        return result

    @staticmethod
    def check_invalid_api(api_trigger: str):
        """
        Makes sure the URL for the API trigger matches the Census format specified below. If it does
        not, it will raise a ValueError.

        Format of api_trigger:
            - https://bearer:secret-token:{secret}@app.getcensus.com/api/v1/syncs/{sync_id}/trigger

        Args:
            - api_trigger (str): If specified in the constructor, will call this validation there

        Returns:
            - confirmed_pattern (Match Object - https://docs.python.org/3/library/re.html#match-objects)
        """
        pattern = r"https:\/\/bearer:secret-token:(.*)@app.getcensus.com\/api\/v1\/syncs\/(\d*)\/trigger"
        url_pattern = re.compile(pattern)
        # Making sure it is a valid api trigger.
        confirmed_pattern = url_pattern.match(api_trigger)

        if not confirmed_pattern:
            raise ValueError(
                """Invalid parameter for `api_trigger` please paste directly from the Census
                                sync configuration page. This is CaSe SenSITiVe."""
            )
        return confirmed_pattern



"""
Tasks for interacting with various Kubernetes API objects.

Note that depending on how you choose to authenticate, tasks in this collection might require
a Prefect Secret called `"KUBERNETES_API_KEY"` that stores your Kubernetes API Key;
this Secret must be a string and in BearerToken format.
"""
try:
    from prefect.tasks.kubernetes.deployment import (
        CreateNamespacedDeployment,
        DeleteNamespacedDeployment,
        ListNamespacedDeployment,
        PatchNamespacedDeployment,
        ReadNamespacedDeployment,
        ReplaceNamespacedDeployment,
    )
    from prefect.tasks.kubernetes.job import (
        CreateNamespacedJob,
        DeleteNamespacedJob,
        ListNamespacedJob,
        PatchNamespacedJob,
        ReadNamespacedJob,
        ReplaceNamespacedJob,
        RunNamespacedJob,
    )
    from prefect.tasks.kubernetes.pod import (
        ConnectGetNamespacedPodExec,
        CreateNamespacedPod,
        DeleteNamespacedPod,
        ListNamespacedPod,
        PatchNamespacedPod,
        ReadNamespacedPod,
        ReplaceNamespacedPod,
        ReadNamespacedPodLogs,
    )
    from prefect.tasks.kubernetes.secrets import KubernetesSecret
    from prefect.tasks.kubernetes.service import (
        CreateNamespacedService,
        DeleteNamespacedService,
        ListNamespacedService,
        PatchNamespacedService,
        ReadNamespacedService,
        ReplaceNamespacedService,
    )
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.kubernetes` requires Prefect to be installed with the "kubernetes" extra.'
    ) from err

__all__ = [
    "ConnectGetNamespacedPodExec",
    "CreateNamespacedDeployment",
    "CreateNamespacedJob",
    "CreateNamespacedPod",
    "CreateNamespacedService",
    "DeleteNamespacedDeployment",
    "DeleteNamespacedJob",
    "DeleteNamespacedPod",
    "DeleteNamespacedService",
    "KubernetesSecret",
    "ListNamespacedDeployment",
    "ListNamespacedJob",
    "ListNamespacedPod",
    "ListNamespacedService",
    "PatchNamespacedDeployment",
    "PatchNamespacedJob",
    "PatchNamespacedPod",
    "PatchNamespacedService",
    "ReadNamespacedDeployment",
    "ReadNamespacedJob",
    "ReadNamespacedPod",
    "ReadNamespacedPodLogs",
    "ReadNamespacedService",
    "ReplaceNamespacedDeployment",
    "ReplaceNamespacedJob",
    "ReplaceNamespacedPod",
    "ReplaceNamespacedService",
    "RunNamespacedJob",
]



"""
This module contains a collection of tasks for interacting with Databricks resources.
"""

try:
    from prefect.tasks.databricks.databricks_submitjob import (
        DatabricksSubmitRun,
        DatabricksRunNow,
        DatabricksSubmitMultitaskRun,
    )
    from prefect.tasks.databricks.databricks_get_job_id import DatabricksGetJobID
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.databricks` requires Prefect to be installed with the "databricks" extra.'
    ) from err

__all__ = [
    "DatabricksRunNow",
    "DatabricksSubmitRun",
    "DatabricksSubmitMultitaskRun",
    "DatabricksGetJobID",
]



"""
This module contains a collection of tasks for interacting with Firebolt databases via
the firebolt-python-sdk library.
"""

try:
    from prefect.tasks.firebolt.firebolt import FireboltQuery
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.firebolt` requires Prefect to be installed with the "firebolt" extra.'
    ) from err



import os
from typing import Optional
from transform import MQLClient
from transform.exceptions import AuthException, QueryRuntimeException, URLException

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from prefect.engine.signals import FAIL


class TransformCreateMaterialization(Task):
    """
    Task to create a materialization against a Transform metrics layer
    deployment.
    Please refer to [Transform official documentation](https://docs.transform.co/)
    for more information.
    This task uses [Transform official MQL Client](https://pypi.org/project/transform/)
    under the hood.

    Args:
        - api_key (str, optional): Transform API Key to be used to
            connect to Transform MQL Server.
        - api_key_env_var (str, optional): The name of the environment variable
            that contains the API Key to be used to connect to Transform MQL Server.
        - mql_server_url (str, optional): The URL of the Transform MQL Server
            from which to create the materialization.
        - mql_server_url_env_var (str, optional): The name of the environment variable
            that contains the URL of the Transform MQL Server from which to
            create the materialization.
        - materialization_name (str, optional): The name of the Transform materialization
            to create.
        - model_key_id (int, optional): The unique identifier of the Transform model
            against which the transformation will be created.
        - start_time (str, optional): The UTC start time of the materialization.
        - end_time (str, optional): The UTC end time of the materialization.
        - output_table (str, optional): The name of the database table, in the form of
            `schema_name.table_name`, where the materialization will be created.
        - force (bool, optional): Whether to force the materialization creation
            or not. Defaults to `False`.
        - wait_for_creation (bool, optional): Whether to wait for the materialization creation
            or not. Defaults to `True`.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        api_key: str = None,
        api_key_env_var: str = None,
        mql_server_url: str = None,
        mql_server_url_env_var: str = None,
        materialization_name: str = None,
        model_key_id: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        output_table: Optional[str] = None,
        force: Optional[bool] = False,
        wait_for_creation: Optional[bool] = True,
        **kwargs,
    ):
        self.api_key = api_key
        self.api_key_env_var = api_key_env_var
        self.mql_server_url = mql_server_url
        self.mql_server_url_env_var = mql_server_url_env_var
        self.materialization_name = materialization_name
        self.model_key_id = model_key_id
        self.start_time = start_time
        self.end_time = end_time
        self.output_table = output_table
        self.force = force
        self.wait_for_creation = wait_for_creation
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "api_key",
        "api_key_env_var",
        "mql_server_url",
        "mql_server_url_env_var",
        "materialization_name",
        "model_key_id",
        "start_time",
        "end_time",
        "output_table",
        "force",
        "wait_for_creation",
    )
    def run(
        self,
        api_key: str = None,
        api_key_env_var: str = None,
        mql_server_url: str = None,
        mql_server_url_env_var: str = None,
        materialization_name: str = None,
        model_key_id: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        output_table: Optional[str] = None,
        force: Optional[bool] = False,
        wait_for_creation: Optional[bool] = True,
    ):
        """
        Task run method to create a materialization against a Transform metrics
        layer deployment.
        All parameters can be provided either during task initialization or directly
        in this `run` method.

        Args:
            - api_key (str, optional): Transform API Key to be used to
                connect to Transform MQL Server.
            - api_key_env_var (str, optional): The name of the environment variable
                that contains the API Key to be used to connect to Transform MQL Server.
            - mql_server_url (str, optional): The URL of the Transform MQL Server
                from which to create the materialization.
            - mql_server_url_env_var (str, optional): The name of the environment variable
                that contains the URL of the Transform MQL Server from which to
                create the materialization.
            - materialization_name (str, optional): The name of the Transform materialization
                to create.
            - model_key_id (int, optional): The unique identifier of the Transform model
                against which the transformation will be created.
            - start_time (str, optional): The UTC start time of the materialization.
            - end_time (str, optional): The ISO end time of the materialization.
            - output_table (str, optional): The name of the database table, in the form of
                `schema_name.table_name`, where the materialization will be created.
            - force (bool, optional): Whether to force the materialization creation
                or not. Defaults to `False`.
            - wait_for_creation (bool, optional): Whether to wait for the materialization creation
                or not. Defaults to `True`.

        Raises:
            - `ValueError` if both `api_key` and `api_key_env_var` are missing.
            - `ValueError` if both `mql_server_url` and `mql_server_url_env_var` are missing.
            - `ValueError` if `materialization_name` is missing.
            - `prefect.engine.signals.FAIL` if the connection with the Transform server cannot
                be established.
            - `prefect.engine.signals.FAIL` if the materialization creation process fails.

        Returns:
            - An `MqlMaterializeResp` object if `wait_for_creation` is `True`.
            - An `MqlQueryStatusResp` object if `wait_for_creation` is `False`.

        """
        # Raise error if both api_key and api_key_env_var are missing
        if not (api_key or api_key_env_var):
            msg = "Both `api_key` and `api_key_env_var` are missing."
            raise ValueError(msg)

        # Raise error if api_key is missing and env var is not found
        if not api_key and api_key_env_var not in os.environ.keys():
            msg = "`api_key` is missing and `api_key_env_var` not found in env vars."
            raise ValueError(msg)

        # Raise error if both mql_server_url and mql_server_url_env_var are missing
        if not (mql_server_url or mql_server_url_env_var):
            msg = "Both `mql_server_url` and `mql_server_url_env_var` are missing."
            raise ValueError(msg)

        # Raise error if mql_server_url is missing and env var is not found
        if not mql_server_url and mql_server_url_env_var not in os.environ.keys():
            msg = "`mql_server_url` is missing and `mql_server_url_env_var` not found in env vars."
            raise ValueError(msg)

        if not materialization_name:
            msg = "`materialization_name` is missing."
            raise ValueError(msg)

        mql_api_key = api_key or os.environ[api_key_env_var]
        mql_url = mql_server_url or os.environ[mql_server_url_env_var]
        use_async = not wait_for_creation

        try:
            mql_client = MQLClient(
                api_key=mql_api_key, mql_server_url=mql_url, use_async=use_async
            )
        except (AuthException, URLException) as e:
            msg = f"Cannot connect to Transform server! Error is: {e.msg}"
            raise FAIL(message=msg)

        response = None
        if use_async:
            response = mql_client.create_materialization(
                materialization_name=materialization_name,
                start_time=start_time,
                end_time=end_time,
                model_key_id=model_key_id,
                output_table=output_table,
                force=force,
            )
            if response.is_failed:
                msg = f"Transform materialization async creation failed! Error is: {response.error}"
                raise FAIL(message=msg)
        else:
            try:
                response = mql_client.materialize(
                    materialization_name=materialization_name,
                    start_time=start_time,
                    end_time=end_time,
                    model_key_id=model_key_id,
                    output_table=output_table,
                    force=force,
                )
            except QueryRuntimeException as e:
                msg = (
                    f"Transform materialization sync creation failed! Error is: {e.msg}"
                )
                raise FAIL(message=msg)

        return response



import json
import logging
import os
import runpy
import sys
import textwrap
import uuid
import time
from contextlib import contextmanager
from types import ModuleType
from typing import Callable, Dict, List, Union, Any

import click
from click import ClickException

import prefect
from prefect.backend.flow import FlowView
from prefect.backend.flow_run import FlowRunView, watch_flow_run
from prefect.backend.execution import execute_flow_run_in_subprocess
from prefect.cli.build_register import (
    TerminalError,
    handle_terminal_error,
    log_exception,
)
from prefect.client import Client
from prefect.utilities.importtools import import_object
from prefect.utilities.logging import temporary_logger_config


@contextmanager
def temporary_environ(environ):
    """
    Temporarily add environment variables to the current os.environ
    The original environment will be restored at context exit
    """
    old_environ = os.environ.copy()
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


@contextmanager
def try_error_done(
    message: str,
    echo: Callable = click.secho,
    traceback: bool = False,
    skip_done: bool = False,
):
    """
    Try to run the code in the context block. On error print "Error" and raise a
    terminal error with the exception string. On succecss, print "Done".

    Args:
        message: The first message to display
        echo: The function to use to echo. Must support `click.secho` arguments
        traceback: Display the exception traceback instead of a short message
        skip_done: Do not display 'Done', the user of the context should instead

    Example:
        >>> with try_error_done("Setting up foo..."):
        >>>    pass
        Setting up foo... Done
        >>> with try_error_done("Setting up bar..."):
        >>>    raise ValueError("no!")
        Setting up bar... Error
        no!
    """
    echo(message, nl=False)
    try:
        yield
    except TerminalError:
        echo(" Error", fg="red")
        raise
    except Exception as exc:
        echo(" Error", fg="red")

        if traceback and not isinstance(exc, TerminalError):
            log_exception(exc, indent=2)
            raise TerminalError
        else:
            raise TerminalError(f"{type(exc).__name__}: {exc}")

    else:
        if not skip_done:
            echo(" Done", fg="green")


def echo_with_log_color(log_level: int, message: str, **kwargs: Any):
    if log_level >= logging.ERROR:
        kwargs.setdefault("fg", "red")
    elif log_level >= logging.WARNING:
        kwargs.setdefault("fg", "yellow")
    elif log_level <= logging.DEBUG:
        kwargs.setdefault("fg", "white")
        kwargs.setdefault("dim", True)
    else:
        kwargs.setdefault("fg", "white")

    click.secho(
        message,
        **kwargs,
    )


def load_flows_from_script(path: str) -> "List[prefect.Flow]":
    """Given a file path, load all flows found in the file"""
    # TODO: This is copied and slightly modified from `prefect.cli.build_register`
    #       we should probably abstract this in the future
    # Temporarily add the flow's local directory to `sys.path` so that local
    # imports work. This ensures that `sys.path` is the same as it would be if
    # the flow script was run directly (i.e. `python path/to/flow.py`).
    orig_sys_path = sys.path.copy()
    sys.path.insert(0, os.path.dirname(os.path.abspath(path)))
    try:
        with prefect.context({"loading_flow": True, "local_script_path": path}):
            namespace = runpy.run_path(path, run_name="<flow>")
    except FileNotFoundError as exc:
        if path in str(exc):  # Only capture it if it's about our file
            raise TerminalError(f"File does not exist: {os.path.abspath(path)!r}")
        raise
    finally:
        sys.path[:] = orig_sys_path

    flows = [f for f in namespace.values() if isinstance(f, prefect.Flow)]
    return flows


def load_flows_from_module(name: str) -> "List[prefect.Flow]":
    """
    Given a module name (or full import path to a flow), load all flows found in the
    module
    """
    # TODO: This is copied and slightly modified from `prefect.cli.build_register`
    #       we should probably abstract this in the future
    try:
        with prefect.context({"loading_flow": True}):
            mod_or_obj = import_object(name)
    except Exception as exc:
        # If the requested module isn't found, log without a traceback
        # otherwise log a general message with the traceback.
        if isinstance(exc, ModuleNotFoundError) and (
            name == exc.name
            or (name.startswith(exc.name) and name[len(exc.name)] == ".")
        ):
            raise TerminalError(str(exc).capitalize())
        elif isinstance(exc, AttributeError):
            raise TerminalError(str(exc).capitalize())
        else:
            raise

    if isinstance(mod_or_obj, ModuleType):
        flows = [f for f in vars(mod_or_obj).values() if isinstance(f, prefect.Flow)]
    elif isinstance(mod_or_obj, prefect.Flow):
        flows = [mod_or_obj]
    else:
        raise TerminalError(
            f"Invalid object of type {type(mod_or_obj).__name__!r} found at {name!r}. "
            f"Expected Module or Flow."
        )

    return flows


def get_flow_from_path_or_module(
    path: str = None, module: str = None, name: str = None
):
    location = path if path is not None else module
    flows = load_flows_from_script(path) if path else load_flows_from_module(module)
    flows_by_name = {flow.name: flow for flow in flows}
    flow_names = ", ".join(map(repr, flows_by_name.keys()))

    if not flows:
        raise TerminalError(f"Found no flows at {location}.")

    if len(flows) > 1 and not name:
        raise TerminalError(
            f"Found multiple flows at {location!r}: {flow_names}\n\n"
            f"Specify a flow name to run."
        )
    if name:
        if name not in flows_by_name:
            raise TerminalError(
                f"Did not find {name!r} in flows at {location}. Found {flow_names}"
            )
        flow = flows_by_name[name]

    else:
        flow = list(flows_by_name.values())[0]

    return flow


def get_flow_view(
    flow_or_group_id: str = None,
    project: str = None,
    name: str = None,
) -> "FlowView":
    if flow_or_group_id:
        # Lookup by flow id then flow group id if that fails
        try:
            flow_view = FlowView.from_flow_id(flow_or_group_id)
        except ValueError:
            pass
        else:
            return flow_view

        try:
            flow_view = FlowView.from_flow_group_id(flow_or_group_id)
        except ValueError:
            pass
        else:
            return flow_view

        # Fall through to failure
        raise TerminalError(
            f"Failed to find flow id or flow group id matching {flow_or_group_id!r}"
        )

    if project:
        if not name:
            raise TerminalError(
                "Missing required option `--name`. Cannot look up a flow by project "
                "without also passing a name."
            )
        return FlowView.from_flow_name(flow_name=name, project_name=project)

    if name:
        # If name wasn't provided for use with another lookup, try a global name search
        return FlowView.from_flow_name(flow_name=name)

    # This line should not be reached
    raise RuntimeError("Failed to find matching case for flow lookup.")


def load_json_key_values(
    cli_input: List[str], display_name: str
) -> Dict[str, Union[dict, str, int]]:
    """
    Parse a list of strings formatted as "key=value" where the value is loaded as JSON.

    We do the best here to display a helpful JSON parsing message, e.g.
    ```
    Error: Failed to parse JSON for parameter 'name' with value

        foo

    JSON Error: Expecting value: line 1 column 1 (char 0)
    Did you forget to include quotes? You may need to escape so your shell does not remove them, e.g. \"
    ```

    Args:
        cli_input: A list of "key=value" strings to parse
        display_name: A name to display in exceptions

    Returns:
        A mapping of keys -> parsed values
    """
    parsed = {}

    def cast_value(value: str) -> Any:
        """Cast the value from a string to a valid JSON type; add quotes for the user
        if necessary
        """
        try:
            return json.loads(value)
        except ValueError as exc:
            if (
                "Extra data" in str(exc) or "Expecting value" in str(exc)
            ) and '"' not in value:
                return cast_value(f'"{value}"')
            raise exc

    for spec in cli_input:
        try:
            key, _, value = spec.partition("=")
        except ValueError:
            raise TerminalError(
                f"Invalid {display_name} option {spec!r}. Expected format 'key=value'."
            )

        try:
            parsed[key] = cast_value(value)
        except ValueError as exc:
            indented_value = textwrap.indent(value, prefix="\t")
            raise TerminalError(
                f"Failed to parse JSON for {display_name} {key!r} with value"
                f"\n\n{indented_value}\n\n"
                f"JSON Error: {exc}"
            )

    return parsed


RUN_EPILOG = """
\bExamples:

\b  Run flow in a script locally

\b    $ prefect run -p hello-world.py

\b  Run flow in a module locally

\b    $ prefect run -m prefect.hello_world

\b  Run flow with a non-default parameter locally

\b    $ prefect run -m prefect.hello_world --param name=Marvin

\b  Run registered flow with the backend by flow name and watch execution

\b    $ prefect run -n "hello-world" --watch

\b  Run registered flow with the backend with custom labels

\b    $ prefect run -n "hello-world" --label example --label hello

\b  Run registered flow with the backend by flow id and exit after creation

\b    $ prefect run -i "9a1cd70c-37d7-4cd4-ab91-d41c2700300d"

\b  Run registered flow and pipe flow run id to another program

\b    $ prefect run -n "hello-world" --quiet | post_run.sh

\b  Run registered flow and execute locally without an agent

\b    $ prefect run -n "hello-world" --execute
"""

FLOW_LOOKUP_MSG = """

Look up a flow to run with one of the following option combinations:
    --id
    --name
    --project and --name
    --path (and --name if there are multiple flows in the script)
    --module (and --name if there are multiple flows in the module)

See `prefect run --help` for more details on the options.
"""


@click.group(invoke_without_command=True, epilog=RUN_EPILOG)
# Flow lookup settings -----------------------------------------------------------------
@click.option(
    "--id",
    "-i",
    "flow_or_group_id",
    help=(
        "The UUID of a flow or flow group to run. If a flow group id is given, "
        "the latest flow id will be used for the run."
    ),
)
@click.option(
    "--project",
    help="The name of the Prefect project containing the flow to run.",
)
@click.option(
    "--path",
    "-p",
    help="The path to a file containing the flow to run.",
)
@click.option(
    "--module",
    "-m",
    help="The python module name containing the flow to run.",
)
@click.option(
    "--name",
    "-n",
    help=(
        "The name of a flow to run from the specified file/module/project. If the "
        "source contains multiple flows, this must be provided. "
    ),
)
# Flow run settings --------------------------------------------------------------------
@click.option(
    "--label",
    "labels",
    help=(
        "A label to add to the flow run. May be passed multiple times to specify "
        "multiple labels. If not passed, the labels from the flow group will be used."
    ),
    multiple=True,
)
@click.option("--run-name", help="A name to assign to the flow run.", default=None)
@click.option(
    "--context",
    "context_vars",
    help=(
        "A key, value pair (key=value) specifying a flow context variable. The value "
        "will be interpreted as JSON. May be passed multiple times to specify multiple "
        "context values. Nested values may be set by passing a dict."
    ),
    multiple=True,
)
@click.option(
    "--param",
    "params",
    help=(
        "A key, value pair (key=value) specifying a flow parameter. The value will be "
        "interpreted as JSON. May be passed multiple times to specify multiple "
        "parameter values."
    ),
    multiple=True,
)
@click.option(
    "--log-level",
    help=(
        "The log level to set for the flow run. If passed, the level must be a valid "
        "Python logging level name. If this option is not passed, the default level "
        "for the flow will be used."
    ),
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    default=None,
)
@click.option(
    "--param-file",
    help=(
        "The path to a JSON file containing parameter keys and values. Any parameters "
        "passed with `--param` will take precedence over these values."
    ),
    default=None,
)
@click.option(
    "--idempotency-key",
    help=(
        "A key to prevent duplicate flow runs. If a flow run has already been started "
        "with the provided value, the command will display information for the "
        "existing run. If using `--execute`, duplicate flow runs will exit with an "
        "error. If not using the backing API, this flag has no effect."
    ),
    default=None,
)
@click.option(
    "--execute",
    help=(
        "Execute the flow run in-process without an agent. If this process exits, the "
        "flow run will be marked as 'Failed'."
    ),
    is_flag=True,
)
@click.option(
    "--schedule",
    "-s",
    help=(
        "Execute the flow run according to the schedule attached to the flow. If this "
        "flag is set, this command will wait between scheduled flow runs. If the flow "
        "has no schedule, this flag will be ignored. If used with a non-local run, an "
        "exception will be thrown."
    ),
    is_flag=True,
)
# Display settings ---------------------------------------------------------------------
@click.option(
    "--quiet",
    "-q",
    help=(
        "Disable verbose messaging about the flow run and just print the flow run id."
    ),
    is_flag=True,
)
@click.option(
    "--no-logs",
    help=(
        "Disable streaming logs from the flow run to this terminal. Only state changes "
        "will be displayed. Only applicable when `--watch` is set."
    ),
    is_flag=True,
)
@click.option(
    "--watch",
    "-w",
    help="Wait for the flow run to finish executing and display status information.",
    is_flag=True,
)
@handle_terminal_error
def run(
    flow_or_group_id,
    project,
    path,
    module,
    name,
    labels,
    context_vars,
    params,
    execute,
    idempotency_key,
    schedule,
    log_level,
    param_file,
    run_name,
    quiet,
    no_logs,
    watch,
):
    """Run a flow"""
    # Define a simple function so we don't have to have a lot of `if not quiet` logic
    quiet_echo = (
        (lambda *_, **__: None)
        if quiet
        else lambda *args, **kwargs: click.secho(*args, **kwargs)
    )

    # Cast labels to a list instead of a tuple so we can extend it
    labels = list(labels)

    # Ensure that the user has not passed conflicting options
    given_lookup_options = {
        key
        for key, option in {
            "--id": flow_or_group_id,
            "--project": project,
            "--path": path,
            "--module": module,
        }.items()
        if option is not None
    }
    # Since `name` can be passed in conjunction with several options and also alone
    # it requires a special case here
    if not given_lookup_options and not name:
        raise ClickException(
            "Received no options to look up the flow." + FLOW_LOOKUP_MSG
        )
    if "--id" in given_lookup_options and name:
        raise ClickException(
            "Received too many options to look up the flow; "
            "cannot specifiy both `--name` and `--id`" + FLOW_LOOKUP_MSG
        )
    if len(given_lookup_options) > 1:
        raise ClickException(
            "Received too many options to look up the flow: "
            f"{', '.join(given_lookup_options)}" + FLOW_LOOKUP_MSG
        )

    # Load parameters and context ------------------------------------------------------
    context_dict = load_json_key_values(context_vars, "context")

    file_params = {}
    if param_file:

        try:
            with open(param_file) as fp:
                file_params = json.load(fp)
        except FileNotFoundError:
            raise TerminalError(
                f"Parameter file does not exist: {os.path.abspath(param_file)!r}"
            )
        except ValueError as exc:
            raise TerminalError(
                f"Failed to parse JSON at {os.path.abspath(param_file)!r}: {exc}"
            )

    cli_params = load_json_key_values(params, "parameter")
    conflicting_keys = set(cli_params.keys()).intersection(file_params.keys())
    if conflicting_keys:
        quiet_echo(
            "The following parameters were specified by file and CLI, the CLI value "
            f"will be used: {conflicting_keys}"
        )
    params_dict = {**file_params, **cli_params}

    # Local flow run -------------------------------------------------------------------

    if path or module:
        # We can load a flow for local execution immediately if given a path or module,
        # otherwise, we'll lookup the flow then pull from storage for a local run
        with try_error_done("Retrieving local flow...", quiet_echo, traceback=True):
            flow = get_flow_from_path_or_module(path=path, module=module, name=name)

        # Set the desired log level
        if no_logs:
            log_level = 100  # CRITICAL is 50 so this should do it

        run_info = ""
        if params_dict:
            run_info += f"└── Parameters: {params_dict}\n"
        if context_dict:
            run_info += f"└── Context: {context_dict}\n"

        if run_info:
            quiet_echo("Configured local flow run")
            quiet_echo(run_info, nl=False)

        quiet_echo("Running flow locally...")
        with temporary_logger_config(
            level=log_level,
            stream_fmt="└── %(asctime)s | %(levelname)-7s | %(message)s",
            stream_datefmt="%H:%M:%S",
        ):
            with prefect.context(**context_dict):
                try:
                    result_state = flow.run(
                        parameters=params_dict, run_on_schedule=schedule
                    )
                except Exception as exc:
                    quiet_echo("Flow runner encountered an exception!")
                    log_exception(exc, indent=2)
                    raise TerminalError("Flow run failed!")

        if result_state.is_failed():
            quiet_echo("Flow run failed!", fg="red")
            sys.exit(1)
        else:
            quiet_echo("Flow run succeeded!", fg="green")

        return

    # Backend flow run -----------------------------------------------------------------

    if schedule:
        raise ClickException("`--schedule` can only be specified for local flow runs")

    client = Client()

    # Validate the flow look up options we've been given and get the flow from the
    # backend
    with try_error_done("Looking up flow metadata...", quiet_echo):
        flow_view = get_flow_view(
            flow_or_group_id=flow_or_group_id,
            project=project,
            name=name,
        )

    if log_level:
        run_config = flow_view.run_config
        if not run_config.env:
            run_config.env = {}
        run_config.env["PREFECT__LOGGING__LEVEL"] = log_level
    else:
        run_config = None

    if execute:
        # Add a random label to prevent an agent from picking up this run
        labels.append(f"agentless-run-{str(uuid.uuid4())[:8]}")

    try:  # Handle keyboard interrupts during creation
        flow_run_id = None

        # Create a flow run in the backend
        with try_error_done(
            f"Creating run for flow {flow_view.name!r}...",
            quiet_echo,
            traceback=True,
            # Display 'Done' manually after querying for data to display so there is not
            # a lag
            skip_done=True,
        ):
            flow_run_id = client.create_flow_run(
                flow_id=flow_view.flow_id,
                parameters=params_dict,
                context=context_dict,
                # If labels is an empty list pass `None` to get defaults
                # https://github.com/PrefectHQ/server/blob/77c301ce0c8deda4f8771f7e9991b25e7911224a/src/prefect_server/api/runs.py#L136
                labels=labels or None,
                run_name=run_name,
                # We only use the run config for setting logging levels right now
                run_config=run_config,
                idempotency_key=idempotency_key,
            )

        if quiet:
            # Just display the flow run id in quiet mode
            click.echo(flow_run_id)
            flow_run = None
        else:
            # Grab information about the flow run (if quiet we can skip this query)
            flow_run = FlowRunView.from_flow_run_id(flow_run_id)
            run_url = client.get_cloud_url("flow-run", flow_run_id)

            # Display "Done" for creating flow run after pulling the info so there
            # isn't a weird lag
            quiet_echo(" Done", fg="green")
            quiet_echo(
                textwrap.dedent(
                    f"""
                        └── Name: {flow_run.name}
                        └── UUID: {flow_run.flow_run_id}
                        └── Labels: {flow_run.labels}
                        └── Parameters: {flow_run.parameters}
                        └── Context: {flow_run.context}
                        └── URL: {run_url}
                        """
                ).strip()
            )

    except KeyboardInterrupt:
        # If the user interrupts here, they will expect the flow run to be cancelled
        quiet_echo("\nKeyboard interrupt detected! Aborting...", fg="yellow")
        if flow_run_id:
            client.cancel_flow_run(flow_run_id=flow_run_id)
            quiet_echo("Cancelled flow run.")
        else:
            # The flow run was not created so we can just exit
            quiet_echo("Aborted.")
        return

    # Handle agentless execution
    if execute:
        quiet_echo("Executing flow run...")
        try:
            with temporary_logger_config(
                level=(
                    100 if no_logs or quiet else log_level
                ),  # Disable logging if asked
                stream_fmt="└── %(asctime)s | %(levelname)-7s | %(message)s",
                stream_datefmt="%H:%M:%S",
            ):
                execute_flow_run_in_subprocess(flow_run_id)
        except KeyboardInterrupt:
            quiet_echo("Keyboard interrupt detected! Aborting...", fg="yellow")
            pass

    elif watch:
        try:
            quiet_echo("Watching flow run execution...")
            for log in watch_flow_run(
                flow_run_id=flow_run_id,
                stream_logs=not no_logs,
            ):
                level_name = logging.getLevelName(log.level)
                timestamp = log.timestamp.in_tz(tz="local")
                echo_with_log_color(
                    log.level,
                    f"└── {timestamp:%H:%M:%S} | {level_name:<7} | {log.message}",
                )

        except KeyboardInterrupt:
            quiet_echo("Keyboard interrupt detected!", fg="yellow")
            try:
                cancel = click.confirm(
                    "On exit, we can leave your flow run executing or cancel it.\n"
                    "Do you want to cancel this flow run?",
                    default=True,
                )
            except click.Abort:
                # A second keyboard interrupt will exit without cancellation
                pass
            else:
                if cancel:
                    client.cancel_flow_run(flow_run_id=flow_run_id)
                    quiet_echo("Cancelled flow run.", fg="green")
                    return

            quiet_echo("Exiting without cancelling flow run!", fg="yellow")
            raise  # Re-raise the interrupt

    else:
        # If not watching or executing, exit without checking state
        return

    # Get the final flow run state
    flow_run = FlowRunView.from_flow_run_id(flow_run_id)

    # Wait for the flow run to be done up to 3 seconds
    elapsed_time = 0
    while not flow_run.state.is_finished() and elapsed_time < 3:
        time.sleep(1)
        elapsed_time += 1
        flow_run = flow_run.get_latest()

    # Display the final state
    if flow_run.state.is_failed():
        quiet_echo("Flow run failed!", fg="red")
        sys.exit(1)
    elif flow_run.state.is_successful():
        quiet_echo("Flow run succeeded!", fg="green")
    else:
        quiet_echo(f"Flow run is in unexpected state: {flow_run.state}", fg="yellow")
        sys.exit(1)



import abc
from prometheus_client import (
    CollectorRegistry,
    Gauge,
    pushadd_to_gateway,
    push_to_gateway,
)
from collections import namedtuple
from typing import Dict, List, Optional
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class _GaugeToGatewayBase(Task):
    @abc.abstractmethod
    def _push(self, **push_args):
        pass

    def __init__(
        self,
        pushgateway_url: str = None,
        counter_name: str = None,
        counter_description: str = None,
        grouping_key: List[str] = None,
        job_name: str = None,
        **kwargs,
    ):
        self.pushgateway_url = pushgateway_url
        self.counter_name = counter_name
        self.counter_description = counter_description
        self.job_name = job_name
        self.grouping_key = grouping_key
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pushgateway_url",
        "counter_name",
        "counter_description",
        "job_name",
        "grouping_key",
    )
    def run(
        self,
        values: List[float],
        labels: List[Dict[str, str]],
        pushgateway_url: str = None,
        counter_name: str = None,
        counter_description: str = None,
        grouping_key: Optional[List[str]] = None,
        job_name: str = None,
    ) -> None:
        """
        Push a Gauge to prometheus [PushGateway](https://prometheus.io/docs/practices/pushing/).
        This will unpack the values and labels to create all the a collector that will be pushed
        to pushgateway.

        The following example is checking the data quality of a dataframe and output the metrics
        for the job to the pushgateway

        example:
        ```
        from typing import Any, Dict, List, Tuple
        import pandas as pd
        from prefect import task, Flow, Parameter
        from prefect.tasks.prometheus import PushAddGaugeToGateway


        @task
        def read_csv(path: str) -> pd.DataFrame:
            return pd.read_csv(path)


        @task
        def check_na(
            df: pd.DataFrame, dataframe_name: str
        ) -> Tuple[List[Dict[str, Any]], List[float]]:
            total_rows = df.shape[0]
            lkeys = []
            lvalues = []
            for c in df.columns:
                na_values = len(df[df[c].isna()][c])
                key = {"df": dataframe_name, "column": c}
                lkeys.append(key)
                lvalues.append(na_values / total_rows)

            return (lkeys, lvalues)


        with Flow("Check_Data") as flow:
            source_path = Parameter(name="sourcePath", required=True)
            name = Parameter(name="name", required=True)
            pushgateway_url = Parameter(
                name="pushgatewayUrl", default="http://localhost:9091"
            )
            df = read_csv(source_path)
            check_result = check_na(df, name)
            push_gateway = PushAddGaugeToGateway()
            push_gateway(
                values=check_result[1],
                labels=check_result[0],
                pushgateway_url=pushgateway_url,
                counter_name="null_values_percentage",
                grouping_key=["df"],
                job_name="check_data",
            )

        flow.run(parameters={"sourcePath": "sample_data.csv", "name": "mySample"})

        ```

        Args:
            - values (List[float]): List of the values to push
            - labels (List[Dict[str, str]]): List of the labels to push attached to the values
            - pushgateway_url (str, optional): Url of the prometheus pushgateway instance
            - counter_name (str, optional): Name of the counter
            - counter_description (str, optional): description of the counter
            - grouping_key (str, optional): List of the key used to calculate the grouping key
            - job_name (str, optional): Name of the job

        Returns:
            - None

        Raises:
            - ValueError: if pushgateway_url or counter_name are empty and values and labels list
                    doesn't have the same length
        """
        if not pushgateway_url:
            raise ValueError("You need to provide the pushgateway_url parameter.")

        if not counter_name:
            raise ValueError("You need to provide the counter_name parameter.")

        if not counter_description:
            counter_description = counter_name

        if len(values) != len(labels):
            raise ValueError(
                f"Values and labels need to have the same length. Values length is {len(values)}, "
                f"labels length is {len(labels)}."
            )

        if len(values) == 0 or len(labels) == 0:
            return

        GroupingValues = namedtuple(
            "GroupingValues", ["grouping_key", "labels_names", "labels", "values"]
        )

        groups = {}
        for l, v in zip(labels, values):

            grouping_key_values = (
                {key: l[key] for key in grouping_key} if grouping_key else {}
            )
            group_key_id = "-".join(grouping_key_values.values())
            group_values = groups.setdefault(
                group_key_id, GroupingValues(grouping_key_values, l.keys(), [], [])
            )
            group_values.labels.append(l)
            group_values.values.append(v)

        for group in groups.values():
            registry = CollectorRegistry()
            g = Gauge(
                counter_name,
                counter_description,
                labelnames=group.labels_names,
                registry=registry,
            )

            for k, v in zip(group.labels, group.values):
                g.labels(**k).set(v)

            self._push(
                gateway=pushgateway_url,
                job=job_name,
                grouping_key=group.grouping_key,
                registry=registry,
            )


class PushGaugeToGateway(_GaugeToGatewayBase):
    """
    Task that allow you to push a Gauge to prometheus [PushGateway]
    (https://prometheus.io/docs/practices/pushing/).This method is using the
    [prometheus_client](https://github.com/prometheus/client_python#exporting-to-a-pushgateway).

    This is a push mode task that it will remove all previous items that match the grouping key.

    Some of the main usage of that task is to allow to push inside of your workflow to prometheus like
    number of rows, quality of the values or anything you want to monitor.

    Args:
        - pushgateway_url (str, optional): Url of the prometheus pushgateway instance
        - counter_name (str, optional): Name of the counter
        - counter_description (str, optional): description of the counter
        - grouping_key (str, optional): List of the key used to calculate the grouping key
        - job_name (str, optional): Name of the job
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor

    """

    def __init__(
        self,
        pushgateway_url: str = None,
        counter_name: str = None,
        counter_description: str = None,
        grouping_key: List[str] = None,
        job_name: str = None,
        **kwargs,
    ):
        super().__init__(
            pushgateway_url=pushgateway_url,
            counter_name=counter_name,
            counter_description=counter_description,
            job_name=job_name,
            grouping_key=grouping_key,
            **kwargs,
        )

    def _push(self, **push_args):
        push_to_gateway(**push_args)


class PushAddGaugeToGateway(_GaugeToGatewayBase):
    """
    Task that allow you to push add a Gauge to prometheus [PushGateway]
    (https://prometheus.io/docs/practices/pushing/). This method is using the
    [prometheus_client](https://github.com/prometheus/client_python#exporting-to-a-pushgateway).

    This is a push add mode task that will add value into the same grouping key.

    Some of the main usage of that task is to allow to push inside of your workflow to prometheus like
    number of rows, quality of the values or anything you want to monitor.

    Args:
        - pushgateway_url (str, optional): Url of the prometheus pushgateway instance
        - counter_name (str, optional): Name of the counter
        - counter_description (str, optional): description of the counter
        - grouping_key (str, optional): List of the key used to calculate the grouping key
        - job_name (str, optional): Name of the job
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor

    """

    def __init__(
        self,
        pushgateway_url: str = None,
        counter_name: str = None,
        counter_description: str = None,
        grouping_key: List[str] = None,
        job_name: str = None,
        **kwargs,
    ):
        super().__init__(
            pushgateway_url=pushgateway_url,
            counter_name=counter_name,
            counter_description=counter_description,
            job_name=job_name,
            grouping_key=grouping_key,
            **kwargs,
        )

    def _push(self, **push_args):
        pushadd_to_gateway(**push_args)


# class PushAddGaugeToGateway(_GaugeToGatewayBase):
#     def _push(self, **push_args):
#         pushadd_to_gateway(**push_args)



"""
The Prefect Storage interface encapsulates logic for storing flows. Each
storage unit is able to store _multiple_ flows (with the constraint of name
uniqueness within a given unit).
"""

from warnings import warn

import prefect
from prefect import config
from prefect.storage.base import Storage
from prefect.storage.azure import Azure
from prefect.storage.bitbucket import Bitbucket
from prefect.storage.codecommit import CodeCommit
from prefect.storage.docker import Docker
from prefect.storage.gcs import GCS
from prefect.storage.github import GitHub
from prefect.storage.gitlab import GitLab
from prefect.storage.local import Local
from prefect.storage.module import Module
from prefect.storage.s3 import S3
from prefect.storage.webhook import Webhook
from prefect.storage.git import Git


def get_default_storage_class() -> type:
    """
    Returns the `Storage` class specified in
    `prefect.config.flows.defaults.storage.default_class`. If the value is a string, it will
    attempt to load the already-imported object. Otherwise, the value is returned.

    Defaults to `Local` if the string config value can not be loaded
    """
    config_value = config.flows.defaults.storage.default_class
    if isinstance(config_value, str):
        try:
            return prefect.utilities.serialization.from_qualified_name(config_value)
        except ValueError:
            warn(
                f"Could not import {config_value}; using prefect.storage.Local instead."
            )
            return Local
    else:
        return config_value


__all__ = [
    "Azure",
    "Bitbucket",
    "CodeCommit",
    "Docker",
    "GCS",
    "Git",
    "GitHub",
    "GitLab",
    "Local",
    "Module",
    "S3",
    "Storage",
    "Webhook",
]



import click
import os
from pathlib import Path
import shutil
import subprocess
import tempfile
import time
import warnings
from pathlib import Path

import yaml

import prefect
from prefect import config
from prefect.utilities.cli import add_options


@click.group(hidden=True)
def server():
    """
    Commands for interacting with the Prefect Core server

    \b
    Usage:
        $ prefect server [COMMAND]

    \b
    Arguments:
        start                   Start Prefect Server using docker-compose
        stop                    Stop Prefect Server by removing all containers and networks
        create-tenant           Creates a new tenant in the server instance
                                    Note: running `start` already creates a default tenant
        config                  Write the configured docker-compose.yml to standard
                                    output for manual deployment

    \b
    Examples:
        $ prefect server start
        ...

        $ prefect server config > docker-compose.yml
    """


COMMON_SERVER_OPTIONS = [
    click.option(
        "--version",
        "-v",
        help="The server image versions to use (for example, '0.1.0' or 'master')",
        hidden=True,
    ),
    click.option(
        "--ui-version",
        "-uv",
        help="The UI image version to use (for example, '0.1.0' or 'master')",
        hidden=True,
    ),
    click.option(
        "--no-upgrade",
        "-n",
        help="Pass this flag to avoid running a database upgrade when the database spins up",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--no-ui",
        "-u",
        help="Pass this flag to avoid starting the UI",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--external-postgres",
        "-ep",
        help="This will remove the Postgres service from the docker-compose setup",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--postgres-url",
        help="Postgres connection url to use. Expected format is "
        "postgres://<username>:<password>@hostname:<port>/<dbname>",
        default=None,
        type=str,
        hidden=True,
    ),
    click.option(
        "--postgres-port",
        help="The port used to serve Postgres. Not valid for external Postgres.",
        default=config.server.database.host_port,
        type=str,
        hidden=True,
    ),
    click.option(
        "--hasura-port",
        help="The port used to serve Hasura",
        default=config.server.hasura.host_port,
        type=str,
        hidden=True,
    ),
    click.option(
        "--graphql-port",
        help="The port used to serve the GraphQL API",
        default=config.server.graphql.host_port,
        type=str,
        hidden=True,
    ),
    click.option(
        "--ui-port",
        help="The port used to serve the UI",
        default=config.server.ui.host_port,
        type=str,
        hidden=True,
    ),
    click.option(
        "--expose",
        help="Make the UI and Core server listen on all interfaces",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--server-port",
        help="The port used to serve the Core server",
        default=config.server.host_port,
        type=str,
        hidden=True,
    ),
    click.option(
        "--no-postgres-port",
        help="Disable port map of Postgres to host. Not valid for external Postgres.",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--no-hasura-port",
        help="Disable port map of Hasura to host",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--no-graphql-port",
        help="Disable port map of the GraphqlAPI to host",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--no-ui-port",
        help="Disable port map of the UI to host",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--no-server-port",
        help="Disable port map of the Core server to host",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--use-volume",
        help="Enable the use of a volume for the postgres service. Not valid for external Postgres.",
        is_flag=True,
        hidden=True,
    ),
    click.option(
        "--volume-path",
        help="A path to use for the postgres volume. Not valid for external Postgres.",
        default=config.server.database.volume_path,
        type=str,
        hidden=True,
    ),
]


def setup_compose_file(
    no_ui=False,
    external_postgres=False,
    no_postgres_port=False,
    no_hasura_port=False,
    no_graphql_port=False,
    no_ui_port=False,
    no_server_port=False,
    use_volume=True,
    temp_dir: str = None,
) -> str:
    # Defaults should be set in the `click` command option, these defaults are to
    # simplify testing

    base_compose_path = Path(__file__).parents[0].joinpath("docker-compose.yml")

    temp_dir = temp_dir or tempfile.gettempdir()
    temp_path = os.path.join(temp_dir, "docker-compose.yml")

    # Copy the docker-compose file to the temp location
    shutil.copy2(base_compose_path, temp_path)

    # Exit early if no changes have been made to the compose file
    if not (
        no_postgres_port
        or external_postgres
        or no_hasura_port
        or no_graphql_port
        or no_ui_port
        or no_server_port
        or not use_volume
        or no_ui
    ):
        return temp_path

    # Otherwise, modify the file
    with open(temp_path, "r") as file:
        compose_yml = yaml.safe_load(file)

        if no_postgres_port:
            del compose_yml["services"]["postgres"]["ports"]

        if no_hasura_port:
            del compose_yml["services"]["hasura"]["ports"]

        if no_graphql_port:
            del compose_yml["services"]["graphql"]["ports"]

        if no_ui_port:
            del compose_yml["services"]["ui"]["ports"]

        if no_server_port:
            del compose_yml["services"]["apollo"]["ports"]

        if not use_volume:
            del compose_yml["services"]["postgres"]["volumes"]

        if no_ui:
            del compose_yml["services"]["ui"]

        if external_postgres:
            del compose_yml["services"]["postgres"]
            compose_yml["services"]["hasura"]["depends_on"].remove("postgres")

    with open(temp_path, "w") as f:
        yaml.safe_dump(compose_yml, f)

    return temp_path


def setup_compose_env(
    version=None,
    ui_version=None,
    no_upgrade=None,
    external_postgres=False,
    postgres_url=None,
    postgres_port=None,
    hasura_port=None,
    graphql_port=None,
    ui_port=None,
    expose=False,
    server_port=None,
    volume_path=None,
):
    # Defaults should be set in the `click` command option, these should _not_ be `None`
    # at runtime.

    # if postgres url is provided explicitly, we're using external postgres
    if postgres_url:
        external_postgres = True

    # Pull current version information
    base_version = prefect.__version__.split("+")
    if len(base_version) > 1:
        default_tag = "master"
    else:
        default_tag = f"core-{base_version[0]}"

    db_connection_url = (
        config.server.database.connection_url if postgres_url is None else postgres_url
    )
    if not external_postgres:
        # replace localhost with postgres to use docker-compose dns
        db_connection_url = db_connection_url.replace("localhost", "postgres")

    PREFECT_ENV = dict(
        DB_CONNECTION_URL=db_connection_url,
        GRAPHQL_HOST_PORT=str(graphql_port),
        UI_HOST_PORT=str(ui_port),
        UI_HOST_IP="0.0.0.0" if expose else config.server.ui.host_ip,
        # Pass the Core version so the Server API can return it
        PREFECT_CORE_VERSION=prefect.__version__,
        # Set the server image tag
        PREFECT_SERVER_TAG=version or default_tag,
    )

    APOLLO_ENV = dict(
        HASURA_API_URL=f"http://hasura:{hasura_port}/v1alpha1/graphql",
        HASURA_WS_URL=f"ws://hasura:{hasura_port}/v1alpha1/graphql",
        PREFECT_API_URL=f"http://graphql:{graphql_port}{config.server.graphql.path}",
        PREFECT_API_HEALTH_URL=f"http://graphql:{graphql_port}/health",
        APOLLO_HOST_PORT=str(server_port),
        APOLLO_HOST_IP="0.0.0.0" if expose else config.server.host_ip,
        PREFECT_SERVER__TELEMETRY__ENABLED=(
            "true" if config.server.telemetry.enabled is True else "false"
        ),
    )

    POSTGRES_ENV = (
        dict(
            POSTGRES_HOST_PORT=str(postgres_port),
            POSTGRES_USER=config.server.database.username,
            POSTGRES_PASSWORD=config.server.database.password,
            POSTGRES_DB=config.server.database.name,
            POSTGRES_DATA_PATH=volume_path,
        )
        if not external_postgres
        else dict()
    )

    UI_ENV = dict(
        APOLLO_URL=config.server.ui.apollo_url, PREFECT_UI_TAG=ui_version or default_tag
    )

    HASURA_ENV = dict(HASURA_HOST_PORT=str(hasura_port))

    env = os.environ.copy()

    # Check for env var / cli conflict
    # TODO: Consider removal in next major version, previously the env would override
    #       the cli without warning

    if "PREFECT_UI_TAG" in env and ui_version and env["PREFECT_UI_TAG"] != ui_version:
        warnings.warn(
            "The UI version has been set in the environment via `PREFECT_UI_TAG` and "
            f"the CLI. The environment variable value (={env['PREFECT_UI_TAG']}) will "
            "be ignored."
        )

    if "PREFECT_SERVER_TAG" in env and version and env["PREFECT_SERVER_TAG"] != version:
        warnings.warn(
            "The version has been set in the environment via `PREFECT_SERVER_TAG` and "
            f"the CLI. The environment variable value (={env['PREFECT_SERVER_TAG']}) "
            "will be ignored."
        )

    if "PREFECT_SERVER_DB_CMD" in env and no_upgrade:
        warnings.warn(
            "The database startup command has been set in the environment via "
            "`PREFECT_SERVER_DB_CMD` and `--no-upgrade` has been passed via CLI. The "
            f"environment variable value (={env['PREFECT_SERVER_DB_CMD']}) will be "
            "ignored."
        )
        env.pop("PREFECT_SERVER_DB_CMD")

    # Allow a more complex database command to be passed via env
    env.setdefault(
        "PREFECT_SERVER_DB_CMD",
        "prefect-server database upgrade -y"
        if not no_upgrade
        else "echo 'DATABASE MIGRATIONS SKIPPED'",
    )

    env.update(**PREFECT_ENV, **APOLLO_ENV, **POSTGRES_ENV, **UI_ENV, **HASURA_ENV)

    return env


def warn_for_postgres_settings_when_using_external_postgres(
    no_postgres_port, postgres_port, use_volume, volume_path
) -> None:
    """
    Issue user warnings if Postgres related settings have changed
    """
    if no_postgres_port:
        warnings.warn(
            "Using external Postgres instance, `--no-postgres-port` flag will be ignored."
        )
    if postgres_port != config.server.database.host_port:
        warnings.warn(
            "Using external Postgres instance, `--postgres-port` flag will be ignored."
        )
    if use_volume:
        warnings.warn(
            "Using external Postgres instance, `--use-volume` flag will be ignored."
        )
    if volume_path != config.server.database.volume_path:
        warnings.warn(
            "Using external Postgres instance, `--volume-path` flag will be ignored."
        )


@server.command(hidden=True, name="config")
@add_options(COMMON_SERVER_OPTIONS)
def config_cmd(
    version,
    ui_version,
    no_upgrade,
    no_ui,
    external_postgres,
    postgres_url,
    postgres_port,
    hasura_port,
    graphql_port,
    ui_port,
    expose,
    server_port,
    no_postgres_port,
    no_hasura_port,
    no_graphql_port,
    no_ui_port,
    no_server_port,
    use_volume,
    volume_path,
):
    """
    This command writes the configured docker-compose.yml file to standard output

    \b
    Options:
        --version, -v       TEXT    The server image versions to use (for example, '0.1.0'
                                    or 'master'). Defaults to `core-a.b.c` where `a.b.c.`
                                    is the version of Prefect Core currently running.
        --ui-version, -uv   TEXT    The UI image version to use (for example, '0.1.0' or
                                    'master'). Defaults to `core-a.b.c` where `a.b.c.` is
                                    the version of Prefect Core currently running.
        --no-upgrade, -n            Flag to avoid running a database upgrade when the
                                    database spins up
        --no-ui, -u                 Flag to avoid starting the UI
        --expose                    Flag to expose the server to external hosts by listening
                                    to 0.0.0.0 instead of localhost.

    \b
        --external-postgres, -ep    Disable the Postgres service, connect to an external one instead
        --postgres-url      TEXT    Postgres connection url to use. Expected format is
                                    postgres://<username>:<password>@hostname:<port>/<dbname>

    \b
        --postgres-port     TEXT    Port used to serve Postgres, defaults to '5432'.
                                    Not valid for external Postgres.
        --hasura-port       TEXT    Port used to serve Hasura, defaults to '3000'
        --graphql-port      TEXT    Port used to serve the GraphQL API, defaults to '4201'
        --ui-port           TEXT    Port used to serve the UI, defaults to '8080'
        --server-port       TEXT    Port used to serve the Core server, defaults to '4200'

    \b
        --no-postgres-port          Disable port map of Postgres to host.
                                    Not valid for external Postgres.
        --no-hasura-port            Disable port map of Hasura to host
        --no-graphql-port           Disable port map of the GraphQL API to host
        --no-ui-port                Disable port map of the UI to host
        --no-server-port            Disable port map of the Core server to host

    \b
        --use-volume                Enable the use of a volume for the Postgres service.
                                    Not valid for external Postgres.
        --volume-path       TEXT    A path to use for the Postgres volume, defaults to
                                    '~/.prefect/pg_data'. Not valid for external Postgres.
    """
    # set external postgres flag if the user has provided `--postgres-url`
    if postgres_url is not None:
        external_postgres = True

    if external_postgres:
        warn_for_postgres_settings_when_using_external_postgres(
            no_postgres_port=no_postgres_port,
            postgres_port=postgres_port,
            use_volume=use_volume,
            volume_path=volume_path,
        )

    compose_path = setup_compose_file(
        no_ui=no_ui,
        external_postgres=external_postgres,
        no_postgres_port=no_postgres_port,
        no_hasura_port=no_hasura_port,
        no_graphql_port=no_graphql_port,
        no_ui_port=no_ui_port,
        no_server_port=no_server_port,
        use_volume=use_volume,
    )

    compose_dir_path = str(Path(compose_path).parent)

    env = setup_compose_env(
        version=version,
        ui_version=ui_version,
        no_upgrade=no_upgrade,
        external_postgres=external_postgres,
        postgres_url=postgres_url,
        postgres_port=postgres_port,
        hasura_port=hasura_port,
        graphql_port=graphql_port,
        ui_port=ui_port,
        expose=expose,
        server_port=server_port,
        volume_path=volume_path,
    )

    subprocess.check_call(["docker-compose", "config"], cwd=compose_dir_path, env=env)


@server.command(hidden=True)
@add_options(COMMON_SERVER_OPTIONS)
@click.option(
    "--detach",
    "-d",
    help="Detached mode. Runs Server containers in the background",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--skip-pull",
    help="Pass this flag to skip pulling new images (if available)",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--hide-welcome",
    help="Pass this flag to hide the ASCII welcome message",
    is_flag=True,
    hidden=True,
)
def start(
    version,
    ui_version,
    skip_pull,
    no_upgrade,
    no_ui,
    external_postgres,
    postgres_url,
    detach,
    postgres_port,
    hasura_port,
    graphql_port,
    ui_port,
    expose,
    server_port,
    no_postgres_port,
    no_hasura_port,
    no_graphql_port,
    no_ui_port,
    no_server_port,
    use_volume,
    volume_path,
    hide_welcome,
):
    """
    This command spins up all infrastructure and services for the Prefect Core server

    \b
    Options:
        --version, -v       TEXT    The server image versions to use (for example, '0.1.0'
                                    or 'master'). Defaults to `core-a.b.c` where `a.b.c.`
                                    is the version of Prefect Core currently running.
        --ui-version, -uv   TEXT    The UI image version to use (for example, '0.1.0' or
                                    'master'). Defaults to `core-a.b.c` where `a.b.c.` is
                                    the version of Prefect Core currently running.
        --no-upgrade, -n            Flag to avoid running a database upgrade when the
                                    database spins up
        --no-ui, -u                 Flag to avoid starting the UI

    \b
        --external-postgres, -ep    Disable the Postgres service, connect to an external one instead
        --postgres-url      TEXT    Postgres connection url to use. Expected format
                                    is postgres://<username>:<password>@hostname:<port>/<dbname>

    \b
        --postgres-port     TEXT    Port used to serve Postgres, defaults to '5432'.
                                    Not valid for external Postgres.
        --hasura-port       TEXT    Port used to serve Hasura, defaults to '3000'
        --graphql-port      TEXT    Port used to serve the GraphQL API, defaults to '4201'
        --ui-port           TEXT    Port used to serve the UI, defaults to '8080'
        --server-port       TEXT    Port used to serve the Core server, defaults to '4200'

    \b
        --no-postgres-port          Disable port map of Postgres to host.
                                    Not valid for external Postgres.
        --no-hasura-port            Disable port map of Hasura to host
        --no-graphql-port           Disable port map of the GraphQL API to host
        --no-ui-port                Disable port map of the UI to host
        --no-server-port            Disable port map of the Core server to host

    \b
        --use-volume                Enable the use of a volume for the Postgres service.
                                    Not valid for external Postgres.
        --volume-path       TEXT    A path to use for the Postgres volume, defaults to
                                    '~/.prefect/pg_data' Not valid for external Postgres.

    \b
        --detach, -d                Detached mode. Runs Server containers in the background
        --skip-pull                 Flag to skip pulling new images (if available)
        --hide-welcome              Flag to hide the ASCII welcome message
    """
    # set external postgres flag if the user has provided `--postgres-url`
    if postgres_url is not None:
        external_postgres = True

    if external_postgres:
        warn_for_postgres_settings_when_using_external_postgres(
            no_postgres_port=no_postgres_port,
            postgres_port=postgres_port,
            use_volume=use_volume,
            volume_path=volume_path,
        )

    compose_path = setup_compose_file(
        no_ui=no_ui,
        external_postgres=external_postgres,
        no_postgres_port=no_postgres_port,
        no_hasura_port=no_hasura_port,
        no_graphql_port=no_graphql_port,
        no_ui_port=no_ui_port,
        no_server_port=no_server_port,
        use_volume=use_volume,
    )

    compose_dir_path = str(Path(compose_path).parent)

    env = setup_compose_env(
        version=version,
        ui_version=ui_version,
        no_upgrade=no_upgrade,
        external_postgres=external_postgres,
        postgres_url=postgres_url,
        postgres_port=postgres_port,
        hasura_port=hasura_port,
        graphql_port=graphql_port,
        ui_port=ui_port,
        expose=expose,
        server_port=server_port,
        volume_path=volume_path,
    )

    proc = None
    try:
        if not skip_pull:
            subprocess.check_call(
                ["docker-compose", "pull"], cwd=compose_dir_path, env=env
            )

        cmd = ["docker-compose", "up"]
        if detach:
            cmd.append("--detach")
        proc = subprocess.Popen(cmd, cwd=compose_dir_path, env=env)
        started = False
        with prefect.utilities.configuration.set_temporary_config(
            {
                "cloud.api": "http://localhost:4200",
                "cloud.graphql": "http://localhost:4200/graphql",
                "backend": "server",
            }
        ):
            while not started:
                try:
                    # Get a client with the correct server port
                    client = prefect.Client(
                        api_server=f"{config.server.host}:{server_port}"
                    )
                    client.graphql("query{hello}", retry_on_api_error=False)
                    started = True
                    # Create a default tenant if no tenant exists
                    if not client.get_available_tenants():
                        client.create_tenant(name="default")
                    if not hide_welcome:
                        print(ascii_welcome(ui_port=str(ui_port)))
                except Exception:
                    time.sleep(0.5)
                    pass
            if detach:
                return
            while True:
                time.sleep(0.5)
    except BaseException:
        click.secho(
            "Exception caught; killing services (press ctrl-C to force)",
            fg="white",
            bg="red",
        )
        subprocess.check_output(
            ["docker-compose", "down"], cwd=compose_dir_path, env=env
        )
        if proc:
            proc.kill()
        raise


def ascii_welcome(ui_port="8080"):
    ui_url = click.style(
        f"http://localhost:{ui_port}", fg="white", bg="blue", bold=True
    )
    docs_url = click.style("https://docs.prefect.io", fg="white", bg="blue", bold=True)

    title = r"""
   _____  _____  ______ ______ ______ _____ _______    _____ ______ _______      ________ _____
  |  __ \|  __ \|  ____|  ____|  ____/ ____|__   __|  / ____|  ____|  __ \ \    / /  ____|  __ \
  | |__) | |__) | |__  | |__  | |__ | |       | |    | (___ | |__  | |__) \ \  / /| |__  | |__) |
  |  ___/|  _  /|  __| |  __| |  __|| |       | |     \___ \|  __| |  _  / \ \/ / |  __| |  _  /
  | |    | | \ \| |____| |    | |___| |____   | |     ____) | |____| | \ \  \  /  | |____| | \ \
  |_|    |_|  \_\______|_|    |______\_____|  |_|    |_____/|______|_|  \_\  \/   |______|_|  \_\

    """

    message = f"""
                                            {click.style('WELCOME TO', fg='blue', bold=True)}
  {click.style(title, bold=True)}
   Visit {ui_url} to get started, or check out the docs at {docs_url}
    """

    return message


@server.command(hidden=True)
def stop():
    """
    This command stops all Prefect Server containers that are connected to the
    `prefect-server` network. Note: This will completely remove the `prefect-server` network
    and all associated containers.
    """
    import docker

    client = docker.APIClient()

    networks = client.networks(names=["prefect-server"])

    if not networks:
        click.echo("No running Prefect Server found")
        return
    network_id = networks[0].get("Id")

    click.echo("Stopping Prefect Server containers and network")
    for container in client.inspect_network(network_id).get("Containers").keys():
        client.disconnect_container_from_network(container, network_id)
        client.remove_container(container, force=True)

    client.remove_network(network_id)
    click.echo("Prefect Server stopped")


@server.command(hidden=True)
@click.option(
    "--name",
    "-n",
    help="The name of a tenant to create",
    hidden=True,
)
@click.option(
    "--slug",
    "-s",
    help="The slug of a tenant to create",
    default=None,
    hidden=True,
)
def create_tenant(name, slug):
    """
    This command creates a tenant for the Prefect Server

    \b
    Options:
        --name, -n       TEXT    The name of a tenant to create
        --slug, -n       TEXT    The slug of a tenant to create (optional)
    """
    client = prefect.Client()
    tenant_id = client.create_tenant(name=name, slug=slug)

    click.secho(f"Tenant created with ID: {tenant_id}", fg="green")



"""
This module contains a collection of tasks to run Data Quality tests using soda-spark library
"""
from warnings import warn

warn(
    f"soda-spark has been deprecated. We cannot guarantee that the SodaSparkScan "
    "task will continue to function properly.",
    DeprecationWarning,
    stacklevel=2,
)

try:
    from prefect.tasks.sodaspark.sodaspark_tasks import SodaSparkScan
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.sodaspark` requires Prefect to be installed with the "sodaspark" extra.'
    ) from err

__all__ = ["SodaSparkScan"]



from typing import Union, Iterable

import yaml

from prefect.run_configs.base import RunConfig
from prefect.utilities.filesystems import parse_path


class KubernetesRun(RunConfig):
    """Configure a flow-run to run as a Kubernetes Job.

    Kubernetes jobs are configured by filling in a job template at runtime. A
    job template can be specified either as a path (to be read in at runtime)
    or an in-memory object (which will be stored along with the flow in Prefect
    Cloud/Server). By default the job template configured on the Agent is used.

    Args:
        - job_template_path (str, optional): Path to a job template to use. If
            a local path (no file scheme, or a `file`/`local` scheme), the job
            template will be loaded on initialization and stored on the
            `KubernetesRun` object as the `job_template` field.  Otherwise the
            job template will be loaded at runtime on the agent.  Supported
            runtime file schemes include (`s3`, `gcs`, and `agent` (for paths
            local to the runtime agent)).
        - job_template (str or dict, optional): An in-memory job template to
            use. Can be either a dict, or a YAML string.
        - image (str, optional): The image to use
        - env (dict, optional): Additional environment variables to set on the job
        - cpu_limit (float or str, optional): The CPU limit to use for the job
        - cpu_request (float or str, optional): The CPU request to use for the job
        - memory_limit (str, optional): The memory limit to use for the job
        - memory_request (str, optional): The memory request to use for the job
        - service_account_name (str, optional): A service account name to use
            for this job. If present, overrides any service account configured
            on the agent or in the job template.
        - image_pull_secrets (list, optional): A list of image pull secrets to
            use for this job. If present, overrides any image pull secrets
            configured on the agent or in the job template.
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work
        - image_pull_policy (str, optional): The imagePullPolicy to use for the job.
            https://kubernetes.io/docs/concepts/configuration/overview/#container-images

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = KubernetesRun()
    ```

    Use a local job template, which is stored along with the Flow in Prefect
    Cloud/Server:

    ```python
    flow.run_config = KubernetesRun(
        job_template_path="my_custom_template.yaml"
    )
    ```

    Use a job template stored in S3, but override the image and CPU limit:

    ```python
    flow.run_config = KubernetesRun(
        job_template_path="s3://example-bucket/my_custom_template.yaml",
        image="example/my-custom-image:latest",
        cpu_limit=2,
    )
    ```

    Use an image not tagged with :latest, and set the image pull policy to `Always`:

    ```python
    flow.run_config = KubernetesRun(
        image="example/my-custom-image:my-tag,
        image_pull_policy="Always"
    )
    ```

    Use a custom `job_template` with a custom label (or any other necessary changes to
    the default).

    Note: you can use the default job template, found at
    `/prefect/src/prefect/agent/kubernetes/job_template.yaml` in the repository,
    as a base to build on. Once a `job_template` is specified, the default is no longer
    used:

    ```python
    flow.run_config = KubernetesRun(
        image="example/my-custom-image:my-tag,
        job_template={
            "apiVersion": "batch/v1",
            "kind": "Job",
            "spec": {
                "template": {
                    "metadata": {
                        "labels": {
                            "my-custom-label": "something"
                        }
                    },
                    "spec": {
                        "containers": [
                            {
                                "name": "flow"
                            }
                        ]
                    }
                }
            }
        }
    )
    ```
    """

    def __init__(
        self,
        *,
        job_template_path: str = None,
        job_template: Union[str, dict] = None,
        image: str = None,
        env: dict = None,
        cpu_limit: Union[float, str] = None,
        cpu_request: Union[float, str] = None,
        memory_limit: str = None,
        memory_request: str = None,
        service_account_name: str = None,
        image_pull_secrets: Iterable[str] = None,
        labels: Iterable[str] = None,
        image_pull_policy: str = None,
    ) -> None:
        super().__init__(env=env, labels=labels)
        if job_template_path is not None and job_template is not None:
            raise ValueError(
                "Cannot provide both `job_template_path` and `job_template`"
            )
        if job_template_path is not None:
            parsed = parse_path(job_template_path)
            if parsed.scheme == "file":
                with open(parsed.path) as f:
                    job_template = yaml.safe_load(f)
                    job_template_path = None
        elif job_template is not None:
            # Normalize job templates to objects rather than str
            if isinstance(job_template, str):
                job_template = yaml.safe_load(job_template)

        assert job_template is None or isinstance(job_template, dict)  # mypy

        if cpu_limit is not None:
            cpu_limit = str(cpu_limit)
        if cpu_request is not None:
            cpu_request = str(cpu_request)

        if image_pull_secrets is not None:
            image_pull_secrets = list(image_pull_secrets)

        image_pull_policies = {"Always", "IfNotPresent", "Never"}
        if (
            image_pull_policy is not None
            and image_pull_policy not in image_pull_policies
        ):
            raise ValueError(
                f"Invalid image_pull_policy {image_pull_policy!r}.  "
                "Expected 'Always', 'IfNotPresent', or 'Never'"
            )

        self.job_template_path = job_template_path
        self.job_template = job_template
        self.image = image
        self.cpu_limit = cpu_limit
        self.cpu_request = cpu_request
        self.memory_limit = memory_limit
        self.memory_request = memory_request
        self.service_account_name = service_account_name
        self.image_pull_secrets = image_pull_secrets
        self.image_pull_policy = image_pull_policy



import click
import os
import shutil
import pendulum
from click.exceptions import Abort
from tabulate import tabulate
from pathlib import Path

from prefect import Client, config
from prefect.exceptions import AuthorizationError, ClientError
from prefect.cli.build_register import handle_terminal_error, TerminalError
from prefect.backend import TenantView


# For deleting authentication tokens which have been replaced with API keys
AUTH_TOKEN_SETTINGS_PATH = Path(f"{config.home_dir}/client").expanduser()


def check_override_auth_token():
    # Exists for purging old tokens only
    if config.cloud.get("auth_token"):
        if os.environ.get("PREFECT__CLOUD__AUTH_TOKEN"):
            click.secho(
                "Auth token has been set in the environment. The CLI cannot be used to "
                "manage your auth. Unset the key with `unset PREFECT__CLOUD__AUTH_TOKEN`",
                fg="red",
            )
        else:
            click.secho(
                "Auth token has been set in the config. The CLI cannot be used to "
                "manage your auth. Remove the key from your configuration file.",
                fg="red",
            )
        raise Abort


def abort_on_config_api_key(message: str = None):
    if config.cloud.get("api_key"):
        # Add a leading space if not null
        message = (" " + message) if message else ""
        raise TerminalError(
            "Your API key is set in the Prefect config instead of with the CLI."
            + message
        )


@click.group(hidden=True)
def auth():
    """
    Handle Prefect Cloud authorization.

    \b
    Usage:
        $ prefect auth [COMMAND]

    \b
    Commands:
        login           Log in to Prefect Cloud
        logout          Log out of Prefect Cloud
        list-tenants    List your available tenants
        switch-tenants  Switch to a different tenant
        create-key      Create an API key
        list-keys       List details of existing API keys
        revoke-key      Delete an API key from the backend

    \bExamples:

    \b  Log in using an existing key

    \b    $ prefect auth login --key MY_KEY

    \b  Log out, removing your current key or token

    \b    $ prefect auth logout

    \b  Switch to another tenant by slug

    \b    $ prefect auth switch-tenants --slug test-person

    \b  Create a new API key that expires at the start of the next year

    \b    $ prefect auth create-key -n marvin --expire 2022-1-1
    """
    if config.backend == "server":
        raise click.UsageError(
            "Prefect Server does not have authentication. Change your backend to "
            "Prefect Cloud with `prefect backend cloud` to log in."
        )


@auth.command(hidden=True)
@click.option(
    "--key",
    "-k",
    help="A Prefect Cloud API key.",
)
@handle_terminal_error
def login(key):
    """
    Login to Prefect Cloud

    Create an API key in the UI then login with it here:

        $ prefect auth login -k YOUR-KEY

    You will be switched to the default tenant associated with the key. After login,
    your available tenants can be seen with `prefect auth list-tenants` and you can
    change the default tenant on this machine using `prefect auth switch-tenants`.

    The given key will be stored on disk for later access. Prefect will default to using
    this key for all interaction with the API but frequently overrides can be passed to
    individual commands or functions. To remove your key from disk, see
    `prefect auth logout`.
    """
    if not key:
        raise TerminalError("You must supply an API key!")

    abort_on_config_api_key(
        "To log in with the CLI, remove the config key `prefect.cloud.api_key`"
    )

    # Ignore any tenant id that has been previously set via login
    client = Client(api_key=key)
    client._tenant_id = None

    try:
        tenant_id = client._get_auth_tenant()
    except AuthorizationError:
        raise TerminalError("Unauthorized. Invalid Prefect Cloud API key.")
    except ClientError:
        raise TerminalError("Error attempting to communicate with Prefect Cloud.")
    else:
        client.tenant_id = tenant_id
        client.save_auth_to_disk()
        tenant = TenantView.from_tenant_id(tenant_id)
        click.secho(
            f"Logged in to Prefect Cloud tenant {tenant.name!r} ({tenant.slug})",
            fg="green",
        )
        return


@auth.command(hidden=True)
@click.option(
    "--token",
    "-t",
    help="Log out from the API token based authentication, ignoring API keys",
    is_flag=True,
)
@handle_terminal_error
def logout(token):
    """
    Log out of Prefect Cloud

    This will remove your cached authentication from disk.
    """

    client = Client()

    # Log out of API keys unless given the token flag
    if client.api_key and not token:

        # Check the source of the API key
        abort_on_config_api_key(
            "To log out, remove the config key `prefect.cloud.api_key`"
        )

        click.confirm(
            "Are you sure you want to log out of Prefect Cloud? "
            "This will remove your API key from this machine.",
            default=False,
            abort=True,
        )

        # Clear the key and tenant id then write to the cache
        client.api_key = ""
        client._tenant_id = ""
        client.save_auth_to_disk()

        click.secho("Logged out of Prefect Cloud", fg="green")

    else:
        raise TerminalError(
            "You are not logged in to Prefect Cloud. "
            "Use `prefect auth login` to log in first."
        )


@auth.command()
def purge_tokens():
    check_override_auth_token()

    if not AUTH_TOKEN_SETTINGS_PATH.exists():
        click.secho(
            "The deprecated authentication tokens settings path "
            f"'{AUTH_TOKEN_SETTINGS_PATH}' has already been removed."
        )

    else:
        confirm = click.confirm(
            "Are you sure you want to delete the deprecated authentication token "
            f"settings folder '{AUTH_TOKEN_SETTINGS_PATH}'?"
        )
        if not confirm:
            print("Aborted!")
            return

        shutil.rmtree(AUTH_TOKEN_SETTINGS_PATH)
        print("Removed!")


@auth.command(hidden=True)
@handle_terminal_error
def list_tenants():
    """
    List available tenants
    """
    client = Client()

    try:
        tenants = client.get_available_tenants()
    except AuthorizationError:
        raise TerminalError(
            "You are not authenticated. Use `prefect auth login` first."
        )

    output = []
    for item in tenants:
        active = None
        if item.id == client.tenant_id:
            active = "*"
        output.append([item.name, item.slug, item.id, active])

    click.echo(
        tabulate(
            output,
            headers=["NAME", "SLUG", "ID", ""],
            tablefmt="plain",
            numalign="left",
            stralign="left",
        )
    )


@auth.command(hidden=True)
@click.option(
    "--id", "-i", required=False, help="A Prefect Cloud tenant id.", hidden=True
)
@click.option(
    "--slug", "-s", required=False, help="A Prefect Cloud tenant slug.", hidden=True
)
@click.option(
    "--default",
    "-d",
    is_flag=True,
    help="Switch to the default tenant for your API key",
)
@handle_terminal_error
def switch_tenants(id, slug, default):
    """
    Switch active tenant

    \b
    Options:
        --id, -i    TEXT    A Prefect Cloud tenant id
        --slug, -s  TEXT    A Prefect Cloud tenant slug
    """

    # If the config specifies a tenant explicitly, it is used before this mechanism
    if config.cloud.get("tenant_id"):
        raise TerminalError(
            "Your tenant id has been set in the Prefect config instead of with the "
            "CLI. To switch tenants with the CLI, remove the config key "
            " `prefect.cloud.tenant_id`"
        )

    client = Client()

    if not client.api_key:
        raise TerminalError("You are not logged in!")

    if default:
        # Clear the set tenant on disk
        client.tenant_id = None
        client.save_auth_to_disk()
        click.secho(
            "Tenant restored to the default tenant for your API key: "
            f"{client._get_auth_tenant()}",
            fg="green",
        )
        return

    try:
        tenant_id = client.switch_tenant(tenant_slug=slug, tenant_id=id)
    except AuthorizationError:
        raise TerminalError("Unauthorized. Your API key is not valid for that tenant.")

    client.save_auth_to_disk()
    click.secho(f"Tenant switched to {tenant_id}", fg="green")


@auth.command(hidden=True)
@click.option("--name", "-n", required=True, help="A name to associate with the key")
@click.option(
    "--expire",
    "-e",
    help=(
        "A optional dateutil parsable time to at. "
        "If not given, the key will never expire."
    ),
    default=None,
)
@click.option(
    "--quiet",
    "-q",
    help="If set, only display the created key.",
    is_flag=True,
)
@handle_terminal_error
def create_key(name, expire, quiet):
    """
    Create a Prefect Cloud API key for authentication with your current user
    """
    # TODO: Add service account associated key creation eventually

    # Parse the input expiration
    if expire is not None:
        try:
            expires_at = pendulum.parse(expire, strict=False)
        except pendulum.parsing.exceptions.ParserError as exc:
            raise TerminalError(
                f"Failed to parse expiration time. {exc}\n"
                "Please pass a date in a dateutil parsable format."
            )

        if expires_at.diff(abs=False).in_seconds() > 0:
            raise TerminalError(
                f"Given expiration time {expire!r} is a time in the past: {expires_at}"
            )
        expire_msg = f" that will expire {expires_at.diff_for_humans()}"
    else:
        expires_at = None
        expire_msg = ""

    client = Client()

    # We must retrieve our own user id first since you could be creating a key for a SA
    if not quiet:
        click.echo("Retrieving user information...")

    response = client.graphql({"query": {"auth_info": {"user_id"}}})
    user_id = response.get("data", {}).get("auth_info", {}).get("user_id")
    if not user_id:
        raise TerminalError("Failed to retrieve the current user id from Prefect Cloud")

    # Actually create the key
    if not quiet:
        click.echo(f"Creating key{expire_msg}...")
    response = client.graphql(
        query={
            "mutation($input: create_api_key_input!)": {
                "create_api_key(input: $input)": {"key"}
            }
        },
        variables=dict(
            input=dict(
                name=name,
                user_id=user_id,
                expires_at=expires_at.in_tz("utc").isoformat() if expires_at else None,
            )
        ),
    )

    key = response.get("data", {}).get("create_api_key", {}).get("key")
    if key is None:
        raise TerminalError(f"Unexpected response from Prefect Cloud: {response}")

    if quiet:
        click.echo(key)
    else:
        click.echo(
            "This is the only time this key will be displayed! Store it somewhere safe."
        )
        click.secho(f"Successfully created key: {key}", fg="green")


@auth.command(hidden=True)
@handle_terminal_error
def list_keys():
    """
    List Prefect Cloud API keys

    If you are a tenant admin, this should list all service account keys as well as keys
    you have created.
    """
    client = Client()

    response = client.graphql(
        query={
            "query": {
                "auth_api_key": {
                    "id": True,
                    "name": True,
                    "expires_at": True,
                }
            }
        }
    )
    keys = response.get("data", {}).get("auth_api_key")
    if keys is None:
        raise TerminalError(f"Unexpected response from Prefect Cloud: {response}")

    if not keys:
        click.secho("You have not created any API keys", fg="yellow")

    else:
        click.echo(
            tabulate(
                [(key.name, key.id, key.expires_at or "NEVER") for key in keys],
                headers=["NAME", "ID", "EXPIRES AT"],
                tablefmt="plain",
                numalign="left",
                stralign="left",
            )
        )


@auth.command(hidden=True)
@click.option(
    "--id",
    "-i",
    required=True,
    help="The UUID for the API key to delete.",
)
@handle_terminal_error
def revoke_key(id):
    """
    Revoke a Prefect Cloud API key.
    """
    client = Client()

    output = client.graphql(
        query={
            "mutation($input: delete_api_key_input!)": {
                "delete_api_key(input: $input)": {"success"}
            }
        },
        variables=dict(input=dict(key_id=id)),
    )

    if not output.get("data", None) or not output.data.delete_api_key.success:
        raise TerminalError(f"Unable to revoke key {id!r}")

    click.secho("Key successfully revoked!", fg="green")


@auth.command(hidden=True)
@handle_terminal_error
def status():
    """
    Get the current Prefect authentication status
    """
    client = Client()
    click.echo(f"You are connecting to {client.api_server}")

    if client.api_key:
        click.echo("You are authenticating with an API key")

        try:
            click.echo(f"You are logged in to tenant {client.tenant_id}")
        except Exception as exc:
            click.echo(f"Your authentication is not working: {exc}")

    if AUTH_TOKEN_SETTINGS_PATH.exists():
        click.secho(
            "The authentication tokens settings path still exists. These have been "
            "removed in favor of API keys. We recommend purging old tokens with "
            "`prefect auth purge-tokens`",
            fg="yellow",
        )

    if config.cloud.get("auth_token"):
        if os.environ.get("PREFECT__CLOUD__AUTH_TOKEN"):
            click.secho(
                "An authentication token is set via environment variable. "
                "These have been removed in favor of API keys and the variable will be "
                "ignored. We recommend unsetting the 'PREFECT__CLOUD__AUTH_TOKEN' key",
                fg="yellow",
            )
        else:
            click.secho(
                "An authentication token is set via the prefect config file. "
                "These have been removed in favor of API keys and the setting will be "
                "ignored. We recommend removing the 'prefect.cloud.auth_token' key",
                fg="yellow",
            )

    if not client.api_key:
        click.secho(
            "You are not logged in! Use `prefect auth login` to login with an API key.",
            fg="yellow",
        )



"""
Tasks for interacting with any SFTP server.
"""
try:
    from prefect.tasks.sftp.sftp import SftpDownload, SftpUpload
except ImportError as err:
    raise ImportError(
        'prefect.tasks.sftp` requires Prefect to be installed with the "sftp" extra.'
    ) from err
